#
# Copyright (c) 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# mirror operations
#

import math

import Generic.point
import Generic.segment
import Generic.circle
import Generic.arc
import Generic.leader
import Generic.polyline
import Generic.hcline
import Generic.vcline
import Generic.acline
import Generic.cline

def _test_point(layer, x, y):
    _p = layer.find('point', x, y)
    if _p is None:
        _p = Generic.point.Point(x, y)
        layer.addObject(_p)
    return _p

def _horizontal_mirror(image, hy, objlist):
    _active_layer = image.getActiveLayer()
    _mobjs = []
    for _obj in objlist:
        if isinstance(_obj, Generic.point.Point):
            _x, _y = _obj.getCoords()
            _ym = _y + (2.0 * (hy - _y))
            _mp = _test_point(_active_layer, _x, _ym)
            _mobjs.append(_mp)
        elif isinstance(_obj, Generic.segment.Segment):
            _p1, _p2 = _obj.getEndpoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()            
            if (abs(hy - _p1y) > 1e-10 or
                abs(hy - _p2y) > 1e-10):
                _y1m = _p1y + (2.0 * (hy - _p1y))
                _mp1 = _test_point(_active_layer, _p1x, _y1m)
                _y2m = _p2y + (2.0 * (hy - _p2y))
                _mp2 = _test_point(_active_layer, _p2x, _y2m)
                _ms = _obj.clone()
                _ms.setP1(_mp1)
                _ms.setP2(_mp2)
                _ms.reset()
                _mobjs.append(_ms)
        elif isinstance(_obj, Generic.arc.Arc):
            _x, _y = _obj.getCenter().getCoords()
            _ym = _y + (2.0 * (hy - _y))
            _mcp = _test_point(_active_layer, _x, _ym)
            _ma = _obj.clone()
            _ma.setCenter(_mcp)
            _ma.setStartAngle(360.0 - _obj.getEndAngle())
            _ma.setEndAngle(360.0 - _obj.getStartAngle())
            _ma.reset()
            _mobjs.append(_ma)
        elif isinstance(_obj, Generic.circle.Circle):
            _x, _y = _obj.getCenter().getCoords()
            if abs(hy - _y) > 1e-10:
                _ym = _y + (2.0 * (hy - _y))
                _mcp = _test_point(_active_layer, _x, _ym)
                _mc = _obj.clone()
                _mc.setCenter(_mcp)
                _mc.reset()
                _mobjs.append(_mc)
        elif isinstance(_obj, Generic.hcline.HCLine):
            _x, _y = _obj.getLocation().getCoords()
            if abs(hy - _y) > 1e-10:
                _ym = _y + (2.0 * (hy - _y))
                _mp = _test_point(_active_layer, _x, _ym)
                _mh = _obj.clone()
                _mh.setLocation(_mp)
                _mh.reset()
                _mobjs.append(_mh)
        elif isinstance(_obj, Generic.vcline.VCLine):
            pass
        elif isinstance(_obj, Generic.acline.ACLine):
            _angle = _obj.getAngle()
            if abs(abs(_angle) - 90.0) > 1e-10: # not vertical
                _x, _y = _obj.getLocation().getCoords()
                _ym = _y + (2.0 * (hy - _y))
                _mp = _test_point(_active_layer, _x, _ym)
                _ma = _obj.clone()
                if abs(_angle) > 1e-10: # not horizontal
                    _ma.setAngle(-1.0 * _angle)
                _ma.setLocation(_mp)
                _ma.reset()
                _mobjs.append(_ma)
        elif isinstance(_obj, Generic.cline.CLine):
            _p1, _p2 = _obj.getKeypoints()
            _x1, _y1 = _p1.getCoords()
            _x2, _y2 = _p2.getCoords()
            if abs(_x2 - _x1) > 1e-10: # not vertical
                _ym = _y1 + (2.0 * (hy - _y1))
                _mp1 = _test_point(_active_layer, _x1, _ym)
                _ym = _y2 + (2.0 * (hy - _y2))
                _mp2 = _test_point(_active_layer, _x2, _ym)
                _mc = Generic.cline.CLine(_mp1, _mp2)
                _mobjs.append(_mc)
        elif isinstance(_obj, Generic.leader.Leader):
            _p1, _p2, _p3 = _obj.getPoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()
            _p3x, _p3y = _p3.getCoords()
            if (abs(_p1y - hy) > 1e-10 or
                abs(_p2y - hy) > 1e-10 or
                abs(_p3y - hy) > 1e-10):
                _ym = _p1y + (2.0 * (hy - _p1y))
                _mp1 = _test_point(_active_layer, _p1x, _ym)
                _ym = _p2y + (2.0 * (hy - _p2y))
                _mp2 = _test_point(_active_layer, _p2x, _ym)
                _ym = _p3y + (2.0 * (hy - _p3y))
                _mp3 = _test_point(_active_layer, _p3x, _ym)
                _ml = _obj.clone()
                _ml.setP1(_mp1)
                _ml.setP2(_mp2)
                _ml.setP3(_mp3)
                _mobjs.append(_ml)
        elif isinstance(_obj, Generic.polyline.Polyline):
            _mpl = _obj.clone()
            for _i in range(len(_mpl)):
                _pt = _mpl.getPoint(_i)
                _x, _y = _pt.getCoords()
                _ym = _y + (2.0 * (hy - _y))
                _mp = _test_point(_active_layer, _x, _ym)
                _mpl.setPoint(_i, _mp)
            _mobjs.append(_mpl)
        else:
            print "skipping obj: " + `_obj`
    for _mobj in _mobjs:
        if not isinstance(_mobj, Generic.point.Point):
            _active_layer.addObject(_mobj)
    return _mobjs

def _vertical_mirror(image, vx, objlist):
    _active_layer = image.getActiveLayer()
    _mobjs = []
    for _obj in objlist:
        if isinstance(_obj, Generic.point.Point):
            _x, _y = _obj.getCoords()
            _xm = _x + (2.0 * (vx - _x))
            _mp = _test_point(_active_layer, _xm, _y)
            _mobjs.append(_mp)
        elif isinstance(_obj, Generic.segment.Segment):
            _p1, _p2 = _obj.getEndpoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()            
            if (abs(vx - _p1x) > 1e-10 or
                abs(vx - _p2x) > 1e-10):
                _x1m = _p1x + (2.0 * (vx - _p1x))
                _mp1 = _test_point(_active_layer, _x1m, _p1y)
                _x2m = _p2x + (2.0 * (vx - _p2x))
                _mp2 = _test_point(_active_layer, _x2m, _p2y)
                _ms = _obj.clone()
                _ms.setP1(_mp1)
                _ms.setP2(_mp2)
                _mobjs.append(_ms)
        elif isinstance(_obj, Generic.arc.Arc):
            _x, _y = _obj.getCenter().getCoords()
            _xm = _x + (2.0 * (vx - _x))
            _mcp = _test_point(_active_layer, _xm, _y)
            _ma = _obj.clone()
            _ma.setCenter(_mcp)
            _ma.setStartAngle(180.0 - _obj.getEndAngle())
            _ma.setEndAngle(180.0 - _obj.getStartAngle())
            _mobjs.append(_ma)
        elif isinstance(_obj, Generic.circle.Circle):
            _x, _y = _obj.getCenter().getCoords()
            if abs(vx - _x) > 1e-10:
                _xm = _x + (2.0 * (vx - _x))
                _mcp = _test_point(_active_layer, _xm, _y)
                _mc = _obj.clone()
                _mc.setCenter(_mcp)
                _mobjs.append(_mc)
        elif isinstance(_obj, Generic.hcline.HCLine):
            pass
        elif isinstance(_obj, Generic.vcline.VCLine):
            _x, _y = _obj.getLocation().getCoords()
            if abs(vx - _x) > 1e-10:
                _xm = _x + (2.0 * (vx - _x))
                _mp = _test_point(_active_layer, _xm, _y)
                _mv = _obj.clone()
                _mv.setLocation(_mp)
                _mobjs.append(_mv)
        elif isinstance(_obj, Generic.acline.ACLine):
            _angle = _obj.getAngle()
            if abs(_angle) > 1e-10: # not horizontal
                _x, _y = _obj.getLocation().getCoords()
                _xm = _x + (2.0 * (vx - _x))
                _mp = _test_point(_active_layer, _xm, _y)
                _ma = _obj.clone()
                if abs(abs(_angle) - 90.0) > 1e-10: # not vertical
                    _ma.setAngle(-1.0 * _angle)
                _ma.setLocation(_mp)
                _mobjs.append(_ma)
        elif isinstance(_obj, Generic.cline.CLine):
            _p1, _p2 = _obj.getKeypoints()
            _x1, _y1 = _p1.getCoords()
            _x2, _y2 = _p2.getCoords()
            if abs(_y2 - _y1) > 1e-10: # not horizontal
                _xm = _x1 + (2.0 * (vx - _x1))
                _mp1 = _test_point(_active_layer, _xm, _y1)
                _xm = _x2 + (2.0 * (vx - _x2))
                _mp2 = _test_point(_active_layer, _xm, _y2)
                _mc = Generic.cline.CLine(_mp1, _mp2)
                _mobjs.append(_mc)
        elif isinstance(_obj, Generic.leader.Leader):
            _p1, _p2, _p3 = _obj.getPoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()
            _p3x, _p3y = _p3.getCoords()
            if (abs(_p1x - vx) > 1e-10 or
                abs(_p2x - vx) > 1e-10 or
                abs(_p3x - vx) > 1e-10):
                _xm = _p1x + (2.0 * (vx - _p1x))
                _mp1 = _test_point(_active_layer, _xm, _p1y)
                _xm = _p2x + (2.0 * (vx - _p2x))
                _mp2 = _test_point(_active_layer, _xm, _p2y)
                _xm = _p3x + (2.0 * (vx - _p3x))
                _mp3 = _test_point(_active_layer, _xm, _p3y)
                _ml = _obj.clone()
                _ml.setP1(_mp1)
                _ml.setP2(_mp2)
                _ml.setP3(_mp3)
                _mobjs.append(_ml)
        elif isinstance(_obj, Generic.polyline.Polyline):
            _mpl = _obj.clone()
            for _i in range(len(_mpl)):
                _pt = _mpl.getPoint(_i)
                _x, _y = _pt.getCoords()
                _xm = _x + (2.0 * (vx - _x))
                _mp = _test_point(_active_layer, _xm, _y)
                _mpl.setPoint(_i, _mp)
            _mobjs.append(_mpl)
        else:
            print "skipping obj: " + `_obj`
    for _mobj in _mobjs:
        if not isinstance(_mobj, Generic.point.Point):
            _active_layer.addObject(_mobj)
    return _mobjs

def _reflect_point(t1, t2, t3, yint, x, y):
    #
    # reflect point in coordinate frame with
    # origin at mirror line y-intersection, then
    # convert reflected point back to global
    # coordinates
    #
    _nx = (t1 * x) + (t3 * (y - yint))
    _ny = (t3 * x) + (t2 * (y - yint)) + yint
    return _nx, _ny

def _angled_mirror(image, slope, yint, objlist):
    _active_layer = image.getActiveLayer()
    _angle = math.atan(slope)
    #
    # linear algebra reflection matrix coefficients
    #
    if abs(abs(_angle) - 45.0) < 1e-10:
        _t1 = 0.0
        _t2 = 0.0
        _t3 = 1.0
    else:
        _cosine = math.cos(_angle)
        _sine = math.sin(_angle)
        _t1 = (2.0 * _cosine * _cosine) - 1.0
        _t2 = (2.0 * _sine * _sine) - 1.0
        _t3 = 2.0 * _cosine * _sine
    _mobjs = []
    for _obj in objlist:
        if isinstance(_obj, Generic.point.Point):
            _x, _y = _obj.getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _np = _test_point(_active_layer, _rx, _ry)
            _mobjs.append(_np)
        elif isinstance(_obj, Generic.segment.Segment):
            _p1, _p2 = _obj.getEndpoints()
            _x, _y = _p1.getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _np1 = _test_point(_active_layer, _rx, _ry)
            _x, _y = _p2.getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _np2 = _test_point(_active_layer, _rx, _ry)
            _ns = _obj.clone()
            _ns.setP1(_np1)
            _ns.setP2(_np2)
            _mobjs.append(_ns)
        elif isinstance(_obj, Generic.arc.Arc):
            _x, _y = _obj.getCenter().getCoords()
            _rcx, _rcy = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _mcp = _test_point(_active_layer, _rcx, _rcy)
            _ep1, _ep2 = _obj.getEndpoints()
            _x, _y = _ep1
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _rtd = 180.0/math.pi
            _ea = _rtd * math.atan2((_ry - _rcy), (_rx - _rcx))
            if _ea < 0.0:
                _ea = _ea + 360.0
            _x, _y = _ep2
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _sa = _rtd * math.atan2((_ry - _rcy), (_rx - _rcx))
            if _sa < 0.0:
                _sa = _sa + 360.0
            _ma = _obj.clone()
            _ma.setCenter(_mcp)
            _ma.setStartAngle(_sa)
            _ma.setEndAngle(_ea)
            _mobjs.append(_ma)
        elif isinstance(_obj, Generic.circle.Circle):
            _x, _y = _obj.getCenter().getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _ncp = _test_point(_active_layer, _rx, _ry)
            _nc = _obj.clone()
            _nc.setCenter(_ncp)
            _mobjs.append(_nc)
        elif isinstance(_obj, Generic.hcline.HCLine):
            _x, _y = _obj.getLocation().getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _mp = _test_point(_active_layer, _rx, _ry)
            if abs(abs(_angle) - 45.0) < 1e-10:
                _mh = Generic.vcline.VCLine(_mp)
            else:
                _xi = (_y - yint)/slope
                _mang = 180.0/math.pi * math.atan2((_ry - _y), (_rx - _xi))
                if _mang < 0.0:
                    _mang = _mang + 360.0
                _mh = Generic.acline.ACLine(_mp, _mang)
            _mobjs.append(_mh)
        elif isinstance(_obj, Generic.vcline.VCLine):
            _x, _y = _obj.getLocation().getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _mp = _test_point(_active_layer, _rx, _ry)
            if abs(abs(_angle) - 45.0) < 1e-10:
                _mv = Generic.hcline.HCLine(_mp)
            else:
                _yi = (slope * _x) + yint 
                _mang = 180.0/math.pi * math.atan2((_ry - _yi), (_rx - _x))
                if _mang < 0.0:
                    _mang = _mang + 360.0
                _mv = Generic.acline.ACLine(_mp, _mang)
            _mobjs.append(_mv)
        elif isinstance(_obj, Generic.acline.ACLine):
            _at45 = False
            if abs(abs(_angle) - 45.0) < 1e-10:
                _at45 = True
            _x, _y = _obj.getLocation().getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _mp = _test_point(_active_layer, _rx, _ry)
            _acl_angle = _obj.getAngle()
            if _at45 and abs(_acl_angle) < 1e-10:
                _ma = Generic.vcline.VCLine(_mp)
            elif _at45 and abs(abs(_acl_angle) - 90.0) < 1e-10:
                _ma = Generic.hcline.HCLine(_mp)
            else:
                if abs(_acl_angle) < 1e-10: # horizontal
                    _xi = (_y - _yint)/slope
                    _yi = _y
                elif abs(abs(_acl_angle) - 90.0) < 1e-10: # vertical
                    _xi = _x
                    _yi = (slope * _x) + _yint
                else:
                    _asl = math.tan((math.pi/180.0) * _acl_angle)
                    if abs(_asl - slope) < 1e-10: # parallel
                        _mang = _acl_angle
                    else:
                        _ayi = _y - (_asl * _x)                        
                        _xi = (_ayi - yint)/(slope - _asl)
                        _yi = (slope * _xi) + yint
                        _mang = 180.0/math.pi * math.atan2((_ry - _yi), (_rx - _xi))
                        if _mang < 0.0:
                            _mang = _mang + 360.0
                _ma = Generic.acline.ACLine(_mp, _mang)
            _mobjs.append(_ma)
        elif isinstance(_obj, Generic.cline.CLine):
            _p1, _p2 = _obj.getKeypoints()
            _x, _y = _p1.getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _mp1 = _test_point(_active_layer, _rx, _ry)
            _x, _y = _p2.getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _mp2 = _test_point(_active_layer, _rx, _ry)
            _nc = Generic.cline.CLine(_mp1, _mp2)
            _mobjs.append(_nc)
        elif isinstance(_obj, Generic.leader.Leader):
            _p1, _p2, _p3 = _obj.getPoints()
            _x, _y = _p1.getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _mp1 = _test_point(_active_layer, _rx, _ry)
            _x, _y = _p2.getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _mp2 = _test_point(_active_layer, _rx, _ry)
            _x, _y = _p3.getCoords()
            _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
            _mp3 = _test_point(_active_layer, _rx, _ry)
            _ml = _obj.clone()
            _ml.setP1(_mp1)
            _ml.setP2(_mp2)
            _ml.setP3(_mp3)
            _mobjs.append(_ml)
        elif isinstance(_obj, Generic.polyline.Polyline):
            _mpts = []
            for _i in range(len(_obj)):
                _x, _y = _obj.getPoint(_i).getCoords()
                _rx, _ry = _reflect_point(_t1, _t2, _t3, yint, _x, _y)
                _mp = _test_point(_active_layer, _rx, _ry)
                _mpts.append(_mp)
            _s = _obj.getStyle()
            _l = _obj.getLinetype()
            _c = _obj.getColor()
            _t = _obj.getThickness()
            _mpl = Generic.polyline.Polyline(_mpts, _s, _l, _c, _t)
            _mobjs.append(_mpl)
        else:
            print "Skipping object: " + `_obj`
    for _mobj in _mobjs:
        if not isinstance(_mobj, Generic.point.Point):
            _active_layer.addObject(_mobj)
    return _mobjs

def mirror_objects(image, mline, objlist):
    _mobjs = []
    if isinstance(mline, Generic.hcline.HCLine):
        _x, _y = mline.getLocation().getCoords()
        _mobjs.extend(_horizontal_mirror(image, _y, objlist))
    elif isinstance(mline, Generic.vcline.VCLine):
        _x, _y = mline.getLocation().getCoords()
        _mobjs.extend(_vertical_mirror(image, _x, objlist))
    elif isinstance(mline, Generic.acline.ACLine):
        _x, _y = mline.getLocation().getCoords()
        _angle = mline.getAngle()
        if abs(_angle) < 1e-10: # horizontal:
            _mobjs.extend(_horizontal_mirror(image, _y, objlist))
        elif abs(abs(_angle) - 90.0) < 1e-10: # vertical
            _mobjs.extend(_vertical_mirror(image, _x, objlist))
        else:
            _slope = math.tan(_angle * (math.pi/180.0))
            _yint = _y - (_slope * _x)
            _mobjs.extend(_angled_mirror(image, _slope, _yint, objlist))
    elif isinstance(mline, Generic.cline.CLine):
        _p1, _p2 = mline.getKeypoints()
        _p1x, _p1y = _p1.getCoords()
        _p2x, _p2y = _p2.getCoords()
        if abs(_p2y - _p1y) < 1e-10: # horizontal
            _mobjs.extend(_horizontal_mirror(image, _p1y, objlist))
        elif abs(_p2x - _p1x) < 1e-10: # vertical
            _mobjs.extend(_vertical_mirror(image, _p1x, objlist))
        else:
            _slope = (_p2y - _p1y)/(_p2x - _p1x)
            _yint = _p1y - (_slope * _p1x)
            _mobjs.extend(_angled_mirror(image, _slope, _yint, objlist))
    else:
        raise TypeError, "Unexpected mirror line type: " + `mline`
    return _mobjs
    



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# construction circle class
#

from __future__ import generators

import math

import point
import conobject
import entity
import tolerance
import quadtree
import logger

class CCircle(conobject.ConstructionObject):
    """A class for contruction circles

A CCircle has two attributes:

center: A Point object
radius: The CCircle's radius

A CCircle has the following methods:

{get/set}Center(): Get/Set the center Point of a CCircle.
{get/set}Radius(): Get/Set the radius of a CCircle.
move(): Move the CCircle.
circumference(): Get the CCircle's circumference.
area(): Get the CCircle's area.
mapCoords(): Find the nearest Point on the CCircle to a coordinate pair.
inRegion(): Returns whether or not a CCircle can be seen in a bounded area.
clone(): Return an indentical copy of a CCircle.
    """

    messages = {
        'moved' : True,
        'center_changed' : True,
        'radius_changed' : True,
        }

    def __init__(self, center, radius, **kw):
        """Initialize a CCircle.

CCircle(center, radius)

The center should be a Point, or a two-entry tuple of floats,
and the radius should be a float greater than 0.
        """
        _cp = center
        if not isinstance(_cp, point.Point):
            _cp = point.Point(center)
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if not _r > 0.0:
            raise ValueError, "Invalid radius: %g" % _r
        conobject.ConstructionObject.__init__(self, **kw)
        self.__radius = _r
        self.__center = _cp
        _cp.connect('moved', self, CCircle.movePoint)
        _cp.storeUser(self)

    def __eq__(self, obj):
        """Compare a CCircle to another for equality.
        """
        if not isinstance(obj, CCircle):
            return False
        if obj is self:
            return True
        _val = False
        if self.__center == obj.getCenter():
            if abs(self.__radius - obj.getRadius()) < 1e-10:
                _val = True
        return _val

    def __ne__(self, obj):
        """Compare a CCircle to another for inequality.
        """
        if not isinstance(obj, CCircle):
            return True
        if obj is self:
            return False
        _val = True
        if self.__center == obj.getCenter():
            if abs(self.__radius - obj.getRadius()) < 1e-10:
                _val = False
        return _val

    def getValues(self):
        _cid = self.__center.getID()
        return 'ccircle', entity.Entity.getValues(self), _cid, self.__radius

    def getCenter(self):
        """Return the center Point of the CCircle.

getCenter()
        """
        return self.__center

    def setCenter(self, c):
        """Set the center Point of the CCircle.

setCenter(c)

The argument must be a Point or a tuple containing
two float values.
        """
        if self.isLocked():
            raise RuntimeError, "Setting center not allowed - object locked."
        _cp = self.__center
        if not isinstance(c, point.Point):
            raise TypeError, "Invalid center point: " + `c`
        if _cp is not c:
            _cp.disconnect(self)
            _cp.freeUser(self)
            self.__center = c
            self.sendMessage('center_changed', _cp)
            c.connect('moved', self, CCircle.movePoint)
            c.storeUser(self)
            if abs(_cp.x - c.x) > 1e-10 or abs(_cp.y - c.y) > 1e-10:
                self.sendMessage('moved', _cp.x, _cp.y, self.__radius)
            self.modified()

    center = property(getCenter, setCenter, None, "CCircle center")

    def getRadius(self):
        """Return the radius of the the CCircle.

getRadius()
        """
        return self.__radius

    def setRadius(self, radius):
        """Set the radius of the CCircle.

setRadius(radius)

The argument must be float value greater than 0.
        """
        if self.isLocked():
            raise RuntimeError, "Setting radius not allowed - object locked."
        _r = radius
        if not isinstance(_r, float):
            _r = float(radius)
        if not _r > 0.0:
            raise ValueError, "Invalid radius: %g" % _r
        _cr = self.__radius
        if abs(_cr - _r) > 1e-10:
            self.__radius = _r
            self.sendMessage('radius_changed', _cr)
            _cx, _cy = self.__center.getCoords()
            self.sendMessage('moved', _cx, _cy, _cr)
            self.modified()

    radius = property(getRadius, setRadius, None, "CCircle radius")

    def move(self, dx, dy):
        """Move a CCircle.

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked():
            raise RuntimeError, "Setting radius not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x, _y = self.__center.getCoords()
            self.ignore('moved')
            self.__center.move(_dx, _dy)
            self.receive('moved')
            self.sendMessage('moved', _x, _y, self.__radius)
            self.modified()

    def circumference(self):
        """Return the circumference of the CCircle.

circumference()
        """
        return 2.0 * math.pi * self.__radius

    def area(self):
        """Return the area enclosed by the CCircle.

area()
        """
        return math.pi * pow(self.__radius, 2)

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the CCircle to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the x-coordinate
y: A Float value giving the y-coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual Point on the CCircle. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _cx, _cy = self.__center.getCoords()
        _r = self.__radius
        _dist = math.hypot((_x - _cx), (_y - _cy))
        if abs(_dist - _r) < _t:
            _angle = math.atan2((_y - _cy),(_x - _cx))
            _xoff = _r * math.cos(_angle)
            _yoff = _r * math.sin(_angle)
            return (_cx + _xoff), (_cy + _yoff)
        return None

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an CCircle exists within a region.

inRegion(xmin, ymin, xmax, ymax[, fully])

The first four arguments define the boundary. The optional
fifth argument "fully" indicates whether or not the CCircle
must be completely contained within the region or just pass
through it.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _xc, _yc = self.__center.getCoords()
        _r = self.__radius
        _val = False
        _bits = 0
        #
        # calculate distances from center to region boundary
        #
        if abs(_xc - _xmin) < _r: _bits = _bits | 1 # left edge
        if abs(_xc - _xmax) < _r: _bits = _bits | 2 # right edge
        if abs(_yc - _ymin) < _r: _bits = _bits | 4 # bottom edge
        if abs(_yc - _ymax) < _r: _bits = _bits | 8 # top edge
        if _bits == 0:
            #
            # if the ccircle center is in region then the entire
            # ccircle is visible since the distance from the center
            # to any edge is greater than the radius. If the center
            # is not in the region then the ccircle is not visible in
            # the region because the distance to any edge is greater
            # than the radius, and so one of the bits should have been
            # set ...
            #
            if ((_xmin < _xc < _xmax) and (_ymin < _yc < _ymax)):
                _val = True
        else:
            _val = True
            #
            # calculate distance to corners of region
            #
            if math.hypot((_xc - _xmin), (_yc - _ymax)) < _r:
                _bits = _bits | 0x10 # upper left
            if math.hypot((_xc - _xmax), (_yc - _ymin)) < _r:
                _bits = _bits | 0x20 # lower right
            if math.hypot((_xc - _xmin), (_yc - _ymin)) < _r:
                _bits = _bits | 0x40 # lower left
            if math.hypot((_xc - _xmax), (_yc - _ymax)) < _r:
                _bits = _bits | 0x80 # upper right
            #
            # if all bits are set then distance from ccircle center
            # to region endpoints is less than radius - ccircle
            # entirely outside the region
            #
            if _bits == 0xff or fully:
                _val = False
        return _val

    def movePoint(self, p, *args):
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _cp = self.__center
        if p is not _cp:
            raise ValueError, "Point is not ccircle center: " + `p`
        _x, _y = _cp.getCoords()
        self.sendMessage('moved', _x, _y, self.__radius)
        self.modified()

    def clone(self):
        """Create an identical copy of a CCircle

clone()
        """
        _cp = self.__center.clone()
        return CCircle(_cp, self.__radius)

    def sendsMessage(self, m):
        if m in CCircle.messages:
            return True
        return conobject.ConstructionObject.sendsMessage(self, m)

#
# Quadtree CCircle storage
#

class CCircleQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def getNodes(self, *args):
        _alen = len(args)
        if _alen != 3:
            raise ValueError, "Expected 3 arguments, got %d" % _alen
        _x, _y, _r = args        
        if not isinstance(_x, float):
            _x = float(args[0])
        if not isinstance(_y, float):
            _y = float(args[1])
        if not isinstance(_r, float):
            _r = float(args[2])
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_cxmin > _xmax) or
                (_cxmax < _xmin) or
                (_cymin > _ymax) or
                (_cymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _cxmax < _xmid: # circle on left side
                    _ne = _se = False
                if _cxmin > _xmid: # circle on right side
                    _nw = _sw = False
                if _cymax < _ymid: # circle below
                    _nw = _ne = False
                if _cymin > _ymid: # circle above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                yield _node

    def addObject(self, obj):
        if not isinstance(obj, CCircle):
            raise TypeError, "Invalid CCircle object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _r = obj.getRadius()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _cxmin - 1.0
            _ymin = _cymin - 1.0
            _xmax = _cxmax + 1.0
            _ymax = _cymax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _cxmin < _xmin:
                _xmin = _cxmin - 1.0
                _resize = True
            if _cxmax > _xmax:
                _xmax = _cxmax + 1.0
                _resize = True
            if _cymin < _ymin:
                _ymin = _cymin - 1.0
                _resize = True
            if _cymax > _ymax:
                _ymax = _cymax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
        for _node in self.getNodes(_x, _y, _r):
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, CCircleQuadtree.moveCCircle)

    def old_addObject(self, obj):
        if not isinstance(obj, CCircle):
            raise TypeError, "Invalid CCircle object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _r = obj.getRadius()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _cxmin - 1.0
            _ymin = _cymin - 1.0
            _xmax = _cxmax + 1.0
            _ymax = _cymax + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _cxmin < _xmin:
                _xmin = _cxmin - 1.0
                _resize = True
            if _cxmax > _xmax:
                _xmax = _cxmax + 1.0
                _resize = True
            if _cymin < _ymin:
                _ymin = _cymin - 1.0
                _resize = True
            if _cymax > _ymax:
                _ymax = _cymax + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = []
        _nodes.append(_node)
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_cxmin > _xmax) or
                (_cxmax < _xmin) or
                (_cymin > _ymax) or
                (_cymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _cxmax < _xmid: # ccircle on left side
                    _ne = _se = False
                if _cxmin > _xmid: # ccircle on right side
                    _nw = _sw = False
                if _cymax < _ymid: # ccircle below
                    _nw = _ne = False
                if _cymin > _ymid: # ccircle above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, CCircleQuadtree.moveCCircle)

    def delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _r = obj.getRadius()
        _pdict = {}
        for _node in self.getNodes(_x, _y, _r):
            _node.delObject(obj) # ccircle may not be in the node ...
            _parent = _node.getParent()
            if _parent is not None:
                _pid = id(_parent)
                if _pid not in _pdict:
                    c_pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def old_delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getCenter().getCoords()
        _r = obj.getRadius()
        _xmin = _ymin = _xmax = _ymax = None
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _pdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_cxmin > _xmax) or
                (_cxmax < _xmin) or
                (_cymin > _ymax) or
                (_cymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _cxmax < _xmid: # ccircle on left side
                    _ne = _se = False
                if _cxmin > _xmid: # ccircle on right side
                    _nw = _sw = False
                if _cymax < _ymid: # ccircle below
                    _nw = _ne = False
                if _cymin > _ymid: # ccircle above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # ccircle may not be in the node ...
                _parent = _node.getParent()
                if _parent is not None:
                    _pid = id(_parent)
                    if _pid not in _pdict:
                        _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def find(self, *args):
        _alen = len(args)
        if _alen < 3:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        _t = tolerance.TOL
        if _alen > 3:
            _t = tolerance.toltest(args[4])
        if not len(self):
            return None
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _cdict = {}
        _bailout = False
        _circ = None
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_cxmin > _xmax) or
                (_cxmax < _xmin) or
                (_cymin > _ymax) or
                (_cymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _cxmax < (_xmid - _t): # ccircle on left side
                    _ne = _se = False
                if _cxmin > (_xmid + _t): # ccircle on right side
                    _nw = _sw = False
                if _cymax < (_ymid - _t): # ccircle below
                    _nw = _ne = False
                if _cymin > (_ymid + _t): # ccircle above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                for _c in _node.getObjects():
                    _cid = id(_c)
                    if _cid not in _cdict:
                        _cx, _cy = _c.getCenter().getCoords()
                        if ((abs(_cx - _x) < _t) and (abs(_cy - _y) < _t)):
                            _cr = _c.getRadius()
                            if abs(_cr - _r) < _t:
                                _bailout = True
                                _circ = _c
                                break
                        _cdict[_cid] = True
            if _bailout:
                break
        return _circ

    def moveCCircle(self, obj, *args):
        if obj not in self:
            raise ValueError, "CCircle not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 3:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        for _node in self.getNodes(_x, _y, _r):
            _node.delObject(obj) # ccircle may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def old_moveCCircle(self, obj, *args):
        if obj not in self:
            raise ValueError, "CCircle not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 3:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _r = args[2]
        if not isinstance(_r, float):
            _r = float(args[2])
        _cxmin = _x - _r
        _cxmax = _x + _r
        _cymin = _y - _r
        _cymax = _y + _r
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_cxmin > _xmax) or
                (_cxmax < _xmin) or
                (_cymin > _ymax) or
                (_cymax < _ymin)):
                continue
            if _node.hasSubnodes():
                _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
                _xmid = (_xmin + _xmax)/2.0
                _ymid = (_ymin + _ymax)/2.0
                _ne = _nw = _sw = _se = True
                if _cxmax < _xmid: # ccircle on left side
                    _ne = _se = False
                if _cxmin > _xmid: # ccircle on right side
                    _nw = _sw = False
                if _cymax < _ymid: # ccircle below
                    _nw = _ne = False
                if _cymin > _ymid: # ccircle above
                    _sw = _se = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # ccircle may not be in node ...
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _circ = _tsep = None
        _bailout = False
        _cdict = {}
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()
            if ((_x < (_xmin - _t)) or
                (_x > (_xmax + _t)) or
                (_y < (_ymin - _t)) or
                (_y > (_ymax + _t))):
                continue
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
            else:
                for _c in _node.getObjects():
                    _cid = id(_c)
                    if _cid not in _cdict:
                        _cp = _c.mapCoords(_x, _y, _t)
                        if _cp is not None:
                            _cx, _cy = _cp
                            _sep = math.hypot((_cx - _x), (_cy - _y))
                            if _tsep is None:
                                _tsep = _sep
                                _circ = _c
                            else:
                                if _sep < _tsep:
                                    _tsep = _sep
                                    _circ = _c
                            if _sep < 1e-10 and _circ is not None:
                                _bailout = True
                                break
            if _bailout:
                break
        return _circ

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _circs = []
        if not len(self):
            return _circs
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _cdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _sxmin, _symin, _sxmax, _symax = _subnode.getBoundary()
                    if ((_sxmin > _xmax) or
                        (_symin > _ymax) or
                        (_sxmax < _xmin) or
                        (_symax < _ymin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _circ in _node.getObjects():
                    _cid = id(_circ)
                    if _cid not in _cdict:
                        if _circ.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _circs.append(_circ)
                        _cdict[_cid] = True
        return _circs

#
# CCircle history class
#

class CCircleLog(logger.Logger):
    def __init__(self, c):
        if not isinstance(c, CCircle):
            raise TypeError, "Invalid circle: " + `c`
        logger.Logger.__init__(self)
        self.__c = c
        c.connect('center_changed' ,self, CCircleLog.centerChange)
        c.connect('radius_changed', self, CCircleLog.radiusChange)

    def detatch(self):
        self.__c.disconnect(self)
        self.__c = None

    def radiusChange(self, c, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _r = args[0]
        if not isinstance(_r, float):
            _r = float(args[0])
        self.saveUndoData('radius_changed', _r)

    def centerChange(self, c, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _old = args[0]
        if not isinstance(_old, point.Point):
            raise TypeError, "Invalid old center point: " + `_old`
        _oid = _old.getID()
        self.saveUndoData('center_changed', _oid)

    def execute(self, undo, *args):
        if undo is not True and undo is not False:
            raise ValueError, "Invalid undo value: " + str(undo)
        _alen = len(args)
        if _alen == 0:
            raise ValueError, "No arguments to execute()"
        _c = self.__c
        _cp = _c.getCenter()
        _op = args[0]
        if _op == 'radius_changed':
            if len(args) < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _r = args[1]
            if not isinstance(_r, float):
                _r = float(args[1])
            _cr = _c.getRadius()
            if undo:
                self.saveUndoData('radius_changed', _cr)
            else:
                self.saveRedoData('radius_changed', _cr)
            self.ignore('radius_changed')
            try:
                if undo:
                    _c.startUndo(True)
                    try:
                        _c.setRadius(_r)
                    finally:
                        _c.endUndo()
                    if abs(_r - _cr) > 1e-10:                        
                        _c.sendMessage('moved', _cp.x, _cp.y, _cr)
                else:
                    _c.setRadius(_r)
            finally:
                self.receive('radius_changed')
        elif _op == 'center_changed':
            if _alen < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _oid = args[1]
            _cid = _cp.getID()
            if undo:
                self.saveRedoData('center_changed', _cid)
                self.ignore('center_changed')
                try:
                    _parent = _c.getParent()
                    if _parent is None:
                        raise ValueError, "CCircle has no parent - cannot undo"
                    _op = _parent.getObject(_oid)
                    if _op is None or not isinstance(_op, point.Point):
                        raise ValueError, "Center point missing: id=%d" % _oid
                    _x1, _y1 = _cp.getCoords()
                    _c.startUndo(True)
                    try:
                        _c.setCenter(_op)
                    finally:
                        _c.endUndo()
                    if (abs(_op.x - _x1) > 1e-10 or
                        abs(_op.y - _y1) > 1e-10):
                        _c.sendMessage('moved', _x1, _y1, _c.getRadius())
                finally:
                    self.receive('center_changed')
            else:
                self.saveUndoData('center_changed', _cid)
                _parent = _c.getParent()
                if _parent is None:
                    raise ValueError, "CCircle has no parent - cannot undo"
                _op = _parent.getObject(_oid)
                if _op is None or not isinstance(_op, point.Point):
                    raise ValueError, "Center point missing: id=%d" % _oid
                self.ignore('center_changed')
                try:
                    _c.setCenter(_op)
                finally:
                    self.receive('center_changed')
        else:
            raise ValueError, "Unexpected operation: %s" % _op



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# GTK Menu building code
#

import os
import sys

import pygtk
pygtk.require('2.0')
import gtk
import gtk.keysyms

import gtkimage
import gtkedit
import gtkentities
import gtkconobjs
import gtkprefs
import gtkdimprefs
import gtkmodify
import gtkmirror
import gtktext
import gtklayers
import gtkdimension
import gtkpycad
import Generic.globals
import Generic.fileio
import Generic.imageio
import Generic.tools
import Generic.menus

#
# dictionary used to map strings to stock icons
#

iconmap = {
    'NEW' : gtk.STOCK_NEW,
    'QUIT' : gtk.STOCK_QUIT,
    'OPEN' : gtk.STOCK_OPEN,
    'CLOSE': gtk.STOCK_CLOSE,
    'DELETE' : gtk.STOCK_DELETE,
    'SAVE' : gtk.STOCK_SAVE,
    'SAVE_AS' : gtk.STOCK_SAVE_AS,
    'UNDO' : gtk.STOCK_UNDO,
    'REDO' : gtk.STOCK_REDO,
    'CUT' : gtk.STOCK_CUT,
    'COPY' : gtk.STOCK_COPY,
    'PASTE': gtk.STOCK_PASTE,
    'PREFS' : gtk.STOCK_PREFERENCES,
    'ZOOM_IN' : gtk.STOCK_ZOOM_IN,
    'ZOOM_OUT' : gtk.STOCK_ZOOM_OUT,
    'ZOOM_FIT' : gtk.STOCK_ZOOM_FIT,
    }

#
# dictionary used to map letters to keysyms
#

accelmap = {
    'N' : gtk.keysyms.N,
    'O' : gtk.keysyms.O,
    'W' : gtk.keysyms.W,
    'S' : gtk.keysyms.S,
    'Q' : gtk.keysyms.Q,
    'Z' : gtk.keysyms.Z,
    'X' : gtk.keysyms.X,
    'C' : gtk.keysyms.C,
    'P' : gtk.keysyms.P,
    'V' : gtk.keysyms.V
    }

#
# callbacks for the menu items
#

def file_menu_init_cb(menuitem, gtkimg):
    # print "in file_menu_init_cb()"
    return gtk.FALSE

def file_new_cb(menuitem, gtkimg):
    _image = gtkimage.GTKImage()
    _background = Generic.globals.prefs['BACKGROUND_COLOR']
    _image.setOption('BACKGROUND_COLOR', _background)
    gtkpycad.add_image(_image)
    _image.window.show_all()
    return gtk.FALSE

def file_open_cb(menuitem, gtkimg):
    _open = False
    _fname = None
    _dialog = gtk.FileSelection("Open File ...")
    _dialog.set_transient_for(gtkimg.getWindow())
    _response = _dialog.run()
    if _response == gtk.RESPONSE_OK:
        _fname = _dialog.get_filename()
        _open = True
    _dialog.destroy()
    if _open:
        _image = gtkimage.GTKImage()
        try:
            _imagefile = Generic.fileio.CompFile(_fname, "r")
            try:
                Generic.imageio.load_image(_image, _imagefile)
            finally:
                _imagefile.close()
        except (IOError, OSError), e:
            sys.stderr.write("Can't open '%s': %s\n" % (_fname, e))
            return gtk.FALSE
        except StandardError, e:
            sys.stderr.write("Error: %s\n" % str(e))
            return gtk.FALSE
        gtkpycad.add_image(_image)
        _image.setFilename(_fname)
        _window = _image.getWindow()
        _window.set_title(os.path.basename(_fname))
        _window.show_all()
        _image.fitImage()
    return gtk.FALSE

def file_close_cb(menuitem, gtkimg):
    gtkpycad.del_image(gtkimg)
    return gtk.FALSE

def _get_filename_and_save(gtkimg, fname=None):
    _window = gtkimg.getWindow()
    _fname = fname
    if _fname is None:
        _fname = _window.get_title()
    _fname = _window.get_title()
    _dialog = gtk.FileSelection("Save As ...")
    _dialog.set_transient_for(gtkimg.getWindow())
    _dialog.set_filename(_fname)
    _response = _dialog.run()
    _save = False
    if _response == gtk.RESPONSE_OK:
        _save = True
        _fname = _dialog.get_filename()
        if _fname == "":
            _fname = 'Untitled.xml'
        if not _fname.endswith('.xml.gz'):
            if not _fname.endswith('.xml'):
                _fname = _fname + '.xml'
        #
        # if the filename already exists see that the user
        # really wants to overwrite it ...
        #
        # test for the filename + '.gz'
        #
        if _fname.endswith('.xml.gz'):
            _gzfile = _fname
        else:
            _gzfile = _fname + '.gz'
        if os.path.exists(_gzfile):
            _save = False
            _dialog2 = gtk.Dialog("Overwrite Existing File", _window,
                                  gtk.DIALOG_MODAL,
                                  (gtk.STOCK_OK, gtk.RESPONSE_OK,
                                   gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
            _hbox = gtk.HBox(gtk.FALSE, 10)
            _hbox.set_border_width(10)
            _dialog2.vbox.pack_start(_hbox, gtk.FALSE, gtk.FALSE, 0)
            _stock = gtk.image_new_from_stock(gtk.STOCK_DIALOG_QUESTION,
                                              gtk.ICON_SIZE_DIALOG)
            _hbox.pack_start(_stock, gtk.FALSE, gtk.FALSE, 0)
            _label = gtk.Label("File already exists. Delete it?")
            _hbox.pack_start(_label, gtk.FALSE, gtk.FALSE, 0)
            _dialog2.show_all()
            _response = _dialog2.run()
            if _response == gtk.RESPONSE_OK:
                _save = True
            _dialog2.destroy()
    _dialog.destroy()
    if _save:
        # print "fname: " + _fname
        gtkimg.setFilename(_fname)
        _window.set_title(os.path.basename(_fname))
        try:
            Generic.imageio.save_image(gtkimg)
        except (IOError, OSError), e:
            sys.stderr.write("Error writing file: %s\n" % str(e))
        except StandardError, e:
            sys.stderr.write("Error: %s\n" % str(e))

def file_save_cb(menuitem, gtkimg):
    _fname = gtkimg.getFilename()
    if _fname is None:
        _get_filename_and_save(gtkimg)
    else:
        try:
            Generic.imageio.save_image(gtkimg)
        except (IOError, OSError), e:
            sys.stderr.write("Error writing file: %s\n" % str(e))
        except StandardError, e:
            sys.stderr.write("Error: %s\n" % str(e))
    return gtk.FALSE

def file_save_as_cb(menuitem, gtkimg):
    _fname = gtkimg.getFilename()
    if _fname is None:
        _fname = gtkimg.getWindow().get_title()
    _get_filename_and_save(gtkimg, _fname)
    return gtk.FALSE

def file_save_layer_cb(menuitem, gtkimage):
    # print "called file_save_layer_cb()"
    active = gtkimage.getActiveLayer()
    layer_name = active.getName()
    dialog = gtk.FileSelection("Save Layer As ...")
    dialog.set_transient_for(gtkimage.getWindow())
    dialog.set_filename(layer_name)
    response = dialog.run()
    if response == gtk.RESPONSE_OK:
        fname = dialog.get_filename()
        print "Saving layer as '%s'" % fname
        #
        # fixme - add the layer saving code ...
        #
    dialog.destroy()
    return gtk.FALSE

def file_quit_cb(menuitem, gtkimage):
    gtk.main_quit()
    return gtk.FALSE

def adjust_edit_menu(menuitem, gtkimage):
    _key = gtkimage.getMenuKey(menuitem)
    if _key is not None:
        if _key == 'edit_undo':
            menuitem.set_sensitive(gtkimage.canUndo())
        elif _key == 'edit_redo':
            menuitem.set_sensitive(gtkimage.canRedo())
        elif _key == 'edit_copy' or _key == 'edit_cut':
            menuitem.set_sensitive(gtkimage.hasSelection())
        elif _key == 'edit_paste':
            menuitem.set_sensitive(Generic.globals.selectobj.hasObjects())
        else:
            pass

def edit_menu_init_cb(menuitem, gtkimage):
    menuitem.get_submenu().foreach(adjust_edit_menu, gtkimage)
    return gtk.FALSE

def edit_undo_cb(menuitem, gtkimage):
    gtkimage.undo()
    gtkimage.redraw()
    return gtk.FALSE

def edit_redo_cb(menuitem, gtkimage):
    gtkimage.redo()
    gtkimage.redraw()
    return gtk.FALSE

def edit_copy_cb(menuitem, gtkimage):
    for lyr, obj in gtkimage.getSelectedObjects():
        Generic.globals.selectobj.storeObject(gtkimage, lyr, obj)
    return gtk.FALSE

def edit_cut_cb(menuitem, gtkimage):
    for lyr, obj in gtkimage.getSelectedObjects():
        Generic.globals.selectobj.storeObject(gtkimage, lyr, obj)
        #
        # check that the object is still in the layer - objects like
        # dimensions get deleted if the points they refer to are
        # deleted ...
        #
        if obj in lyr:
            gtkimage.delObject(obj,lyr)
    gtkimage.redraw()
    return gtk.FALSE

def edit_paste_cb(menuitem, gtkimage):
    _tool = Generic.tools.Tool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click where you want to paste the objects")
    gtkedit.paste_mode_init(_tool)
    return gtk.FALSE

def select_cb(menuitem, gtkimage):
    _tool = Generic.tools.Tool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the items you want to select.")
    gtkedit.select_mode_init(_tool)
    return gtk.FALSE

def select_all_points_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _point in _active_layer.getLayerEntities("point"):
        gtkimage.selectObject(_active_layer, _point)
    return gtk.FALSE

def select_all_segments_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _segment in _active_layer.getLayerEntities("segment"):
        gtkimage.selectObject(_active_layer, _segment)
    return gtk.FALSE

def select_all_circles_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _circle in _active_layer.getLayerEntities("circle"):
        gtkimage.selectObject(_active_layer, _circle)
    return gtk.FALSE

def select_all_arcs_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _arc in _active_layer.getLayerEntities("arc"):
        gtkimage.selectObject(_active_layer, _arc)
    return gtk.FALSE

def select_all_hclines_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _hcl in _active_layer.getLayerEntities("hcline"):
        gtkimage.selectObject(_active_layer, _hcl)
    return gtk.FALSE

def select_all_vclines_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _vcl in _active_layer.getLayerEntities("vcline"):
        gtkimage.selectObject(_active_layer, _vcl)
    return gtk.FALSE

def select_all_aclines_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _acl in _active_layer.getLayerEntities("acline"):
        gtkimage.selectObject(_active_layer, _acl)
    return gtk.FALSE

def select_all_clines_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _cline in _active_layer.getLayerEntities("cline"):
        gtkimage.selectObject(_active_layer, _cline)
    return gtk.FALSE

def select_all_ccircles_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _ccircle in _active_layer.getLayerEntities("ccircle"):
        gtkimage.selectObject(_active_layer, _ccircle)
    return gtk.FALSE

def select_all_leaders_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _leader in _active_layer.getLayerEntities("leader"):
        gtkimage.selectObject(_active_layer, _leader)
    return gtk.FALSE

def select_all_ldims_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _ldim in _active_layer.getLayerEntities("linear_dimension"):
        gtkimage.selectObject(_active_layer, _ldim)
    return gtk.FALSE

def select_all_hdims_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _hdim in _active_layer.getLayerEntities("horizontal_dimension"):
        gtkimage.selectObject(_active_layer, _hdim)
    return gtk.FALSE

def select_all_vdims_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _vdim in _active_layer.getLayerEntities("vertical_dimension"):
        gtkimage.selectObject(_active_layer, _vdim)
    return gtk.FALSE

def select_all_rdims_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _rdim in _active_layer.getLayerEntities("radial_dimension"):
        gtkimage.selectObject(_active_layer, _rdim)
    return gtk.FALSE

def select_all_adims_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _adim in _active_layer.getLayerEntities("angular_dimension"):
        gtkimage.selectObject(_active_layer, _adim)
    return gtk.FALSE

def prefs_cb(menuitem, gtkimage):
    gtkprefs.prefs_dialog(gtkimage)
    return gtk.FALSE

def draw_menu_init_cb(menuitem, gtkimage):
    # print "in draw_menu_init_cb()"
    return gtk.FALSE

def draw_point_cb(menuitem, gtkimage):
    _tool = Generic.tools.PointTool()
    gtkimage.setTool(_tool)
    gtkentities.point_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_segment_cb(menuitem, gtkimage):
    _tool = Generic.tools.SegmentTool()
    gtkimage.setTool(_tool)
    gtkentities.segment_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_rectangle_cb(menuitem, gtkimage):
    _tool = Generic.tools.RectangleTool()
    gtkimage.setTool(_tool)
    gtkentities.rectangle_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_circle_center_cb(menuitem, gtkimage):
    _tool = Generic.tools.CircleTool()
    gtkimage.setTool(_tool)
    gtkentities.circle_center_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_circle_tp_cb(menuitem, gtkimage):
    _tool = Generic.tools.TwoPointCircleTool()
    gtkimage.setTool(_tool)
    gtkentities.circle_tp_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_arc_center_cb(menuitem, gtkimage):
    _tool = Generic.tools.ArcTool()
    gtkimage.setTool(_tool)
    gtkentities.arc_center_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_hcl_cb(menuitem, gtkimage):
    _tool = Generic.tools.HCLineTool()
    gtkimage.setTool(_tool)
    gtkconobjs.hcline_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_vcl_cb(menuitem, gtkimage):
    _tool = Generic.tools.VCLineTool()
    gtkimage.setTool(_tool)
    gtkconobjs.vcline_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_acl_cb(menuitem, gtkimage):
    _tool = Generic.tools.ACLineTool()
    gtkimage.setTool(_tool)
    gtkconobjs.acline_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_cl_cb(menuitem, gtkimage):
    _tool = Generic.tools.CLineTool()
    gtkimage.setTool(_tool)
    gtkconobjs.cline_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_perpendicular_cline_cb(menuitem, gtkimage):
    _tool = Generic.tools.Tool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the object you want a perpendicular to")
    gtkconobjs.perpendicular_cline_mode_init(_tool)
    return gtk.FALSE

def draw_tangent_cline_cb(menuitem, gtkimage):
    _tool = Generic.tools.Tool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the circle object you want a tangent to")
    gtkconobjs.tangent_cline_mode_init(_tool)
    return gtk.FALSE

def draw_tangent_two_ccircles_cb(menuitem, gtkimage):
    _tool = Generic.tools.CCircleTangentLineTool()
    gtkimage.setTool(_tool)
    gtkconobjs.two_circle_tangent_line_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_poffset_cline_cb(menuitem, gtkimage):
    _tool = Generic.tools.ParallelOffsetTool()
    gtkimage.setTool(_tool)
    gtkconobjs.parallel_offset_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_ccirc_cp_cb(menuitem, gtkimage):
    _tool = Generic.tools.CCircleTool()
    gtkimage.setTool(_tool)
    # gtkimage.setPrompt("Click in the drawing area or enter a Point")
    gtkconobjs.ccircle_cpmode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_ccirc_tp_cb(menuitem, gtkimage):
    _tool = Generic.tools.TwoPointCCircleTool()
    gtkimage.setTool(_tool)
    # gtkimage.setPrompt("Click in the drawing area or enter a Point")
    gtkconobjs.ccircle_tpmode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_tangent_ccircle_menu_init_cb(menuitem, gtkimage):
    # print "in draw_tangent_ccircle_menu_init_cb()"
    return gtk.FALSE

def draw_tangent_single_conobj_cb(menuitem, gtkimage):
    _tool = Generic.tools.TangentCCircleTool()
    gtkimage.setTool(_tool)
    gtkconobjs.tangent_ccircle_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_tangent_two_conobjs_cb(menuitem, gtkimage):
    # print "two_conobjs_cb() ..."
    _tool = Generic.tools.TwoPointTangentCCircleTool()
    gtkimage.setTool(_tool)
    gtkconobjs.two_cline_tancc_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_chamfer_cb(menuitem, gtkimage):
    _tool = Generic.tools.Tool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the points where you want a chamfer.")
    gtkentities.chamfer_mode_init(_tool)
    return gtk.FALSE

def draw_fillet_cb(menuitem, gtkimage):
    _tool = Generic.tools.Tool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the points where you want a fillet.")
    gtkentities.fillet_mode_init(_tool)
    return gtk.FALSE

def draw_leader_cb(menuitem, gtkimage):
    _tool = Generic.tools.LeaderTool()
    gtkimage.setTool(_tool)
    gtkentities.leader_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_polyline_cb(menuitem, gtkimage):
    _tool = Generic.tools.PolylineTool()
    gtkimage.setTool(_tool)
    gtkentities.polyline_mode_init(gtkimage, _tool)
    return gtk.FALSE

def _get_polygon_side_count(gtkimg):
    _sides = 0
    _window = gtkimg.getWindow()
    _dialog = gtk.Dialog("Polygon Sides", _window,
                         gtk.DIALOG_MODAL,
                         (gtk.STOCK_OK, gtk.RESPONSE_OK,
                          gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
    _hbox = gtk.HBox(gtk.FALSE, 10)
    _hbox.set_border_width(10)
    _dialog.vbox.pack_start(_hbox, gtk.FALSE, gtk.FALSE, 0)
    _stock = gtk.image_new_from_stock(gtk.STOCK_DIALOG_QUESTION,
                                      gtk.ICON_SIZE_DIALOG)
    _hbox.pack_start(_stock, gtk.FALSE, gtk.FALSE, 0)
    _label = gtk.Label("Number of sides:")
    _hbox.pack_start(_label, gtk.FALSE, gtk.FALSE, 0)
    _adj = gtk.Adjustment(3, 3, 3600, 1, 1, 1) # arbitrary max ...
    _sb = gtk.SpinButton(_adj)
    _sb.set_numeric(gtk.TRUE)
    _hbox.pack_start(_sb, gtk.TRUE, gtk.TRUE, 0)
    _dialog.show_all()
    _response = _dialog.run()
    if _response == gtk.RESPONSE_OK:
        _sides = _sb.get_value()
    _dialog.destroy()
    return _sides

def draw_polygon_cb(menuitem, gtkimage):
    _sides = _get_polygon_side_count(gtkimage)
    if _sides > 0:
        _tool = Generic.tools.PolygonTool()
        _tool.setSideCount(_sides)
        gtkimage.setTool(_tool)
        gtkentities.polygon_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_ext_polygon_cb(menuitem, gtkimage):
    _sides = _get_polygon_side_count(gtkimage)
    if _sides > 0:
        _tool = Generic.tools.PolygonTool()
        _tool.setExternal()
        _tool.setSideCount(_sides)
        gtkimage.setTool(_tool)
        gtkentities.polygon_mode_init(gtkimage, _tool)
    return gtk.FALSE

def draw_set_style(menuitem, gtkimage):
    gtkentities.set_active_style(gtkimage)
    return gtk.FALSE

def draw_set_linetype(menuitem, gtkimage):
    gtkentities.set_active_linetype(gtkimage)
    return gtk.FALSE

def draw_set_color(menuitem, gtkimage):
    gtkentities.set_active_color(gtkimage)
    return gtk.FALSE

def draw_set_thickness(menuitem, gtkimage):
    gtkentities.set_line_thickness(gtkimage)
    return gtk.FALSE

def draw_text_cb(menuitem, gtkimage):
    _tool = Generic.tools.TextTool()
    gtkimage.setTool(_tool)
    gtktext.get_new_text(gtkimage)
    if _tool.hasText():
        gtktext.text_add_init(gtkimage, _tool)
    return gtk.FALSE

def modify_menu_init_cb(menuitem, gtkimage):
    # print "in modify_menu_init_cb()"
    return gtk.FALSE

def move_horizontal_cb(menuitem, gtkimage):
    _tool = Generic.tools.HorizontalMoveTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click in the drawing area or enter the distance.")
    gtkmodify.move_horizontal_init(_tool)
    return gtk.FALSE

def move_vertical_cb(menuitem, gtkimage):
    _tool = Generic.tools.VerticalMoveTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click in the drawing area or enter the distance.")
    gtkmodify.move_vertical_init(_tool)
    return gtk.FALSE

def move_twopoint_cb(menuitem, gtkimage):
    _tool = Generic.tools.MoveTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click in the drawing area.")
    gtkmodify.move_twopoint_init(_tool)
    return gtk.FALSE

def stretch_horiz_cb(menuitem, gtkimage):
    _tool = Generic.tools.HorizontalStretchTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on a point or enter the horizontal distance.")
    gtkmodify.stretch_horizontal_init(_tool)
    return gtk.FALSE

def stretch_vert_cb(menuitem, gtkimage):
    _tool = Generic.tools.VerticalStretchTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click in the drawing area or enter the distance.")
    gtkmodify.stretch_vertical_init(_tool)
    return gtk.FALSE

def stretch_twopoint_cb(menuitem, gtkimage):
    _tool = Generic.tools.StretchTool()
    gtkimage.setTool(_tool)
    print "in stretch_twopoint_cb()" # fixme
    return gtk.FALSE

def transfer_object_cb(menuitem, gtkimage):
    _tool = Generic.tools.Tool()
    gtkimage.setTool(_tool)
    prompt = "Click on the objects you want to transfer to the active layer"
    gtkimage.setPrompt(prompt)
    gtkmodify.transfer_object_init(_tool)
    return gtk.FALSE

def split_object_cb(menuitem, gtkimage):
    _tool = Generic.tools.SplitTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the objects you want to split")
    gtkmodify.split_object_init(_tool)
    return gtk.FALSE

def mirror_object_cb(menuitem, gtkimage):
    _tool = Generic.tools.MirrorTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the mirroring construction line.")
    gtkmirror.mirror_mode_init(_tool)
    return gtk.FALSE

def delete_cb(menuitem, gtkimage):
    _tool = Generic.tools.DeleteTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the items you want to delete.")
    gtkmodify.delete_mode_init(_tool)
    return gtk.FALSE

def change_style_cb(menuitem, gtkimage):
    gtkmodify.change_style_init(gtkimage)
    return gtk.FALSE

def change_color_cb(menuitem, gtkimage):
    gtkmodify.change_color_init(gtkimage)
    return gtk.FALSE

def change_linetype_cb(menuitem, gtkimage):
    gtkmodify.change_linetype_init(gtkimage)
    return gtk.FALSE

def change_thickness_cb(menuitem, gtkimage):
    gtkmodify.change_thickness_init(gtkimage)
    return gtk.FALSE

def zoom_cb(menuitem, gtkimage):
    # print "in zoom_cb()"
    _tool = Generic.tools.Tool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click in the window.")
    gtkmodify.zoom_init(_tool)
    return gtk.FALSE

def zoom_in_cb(menuitem, gtkimage):
    _xmin, _ymin, _xmax, _ymax = gtkimage.getView()
    _scale = gtkimage.getUnitsPerPixel()
    _xdiff = abs(_xmax - _xmin)
    _ydiff = abs(_ymax - _ymin)
    _xmin = (_xmin + _xmax)/2.0 - _xdiff/4.0
    _ymin = (_ymin + _ymax)/2.0 - _ydiff/4.0
    gtkimage.setView(_xmin, _ymin, (_scale/2.0))
    gtkimage.redraw()
    return gtk.FALSE

def zoom_out_cb(menuitem, gtkimage):
    _xmin, _ymin, _xmax, _ymax = gtkimage.getView()
    _scale = gtkimage.getUnitsPerPixel()
    _xdiff = abs(_xmax - _xmin)
    _ydiff = abs(_ymax - _ymin)
    _xmin = (_xmin + _xmax)/2.0 - _xdiff
    _ymin = (_ymin + _ymax)/2.0 - _ydiff
    gtkimage.setView(_xmin, _ymin, (_scale * 2.0))
    gtkimage.redraw()
    return gtk.FALSE

def zoom_fit_cb(menuitem, gtkimage):
    gtkimage.fitImage()
    return gtk.FALSE

def dimension_linear_cb(menuitem, gtkimage):
    _tool = Generic.tools.LinearDimensionTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the first point for the dimension.")
    gtkdimension.linear_mode_init(_tool)
    return gtk.FALSE

def dimension_menu_init_cb(menuitem, gtkimg):
    # print "in dimension_menu_init_cb()"
    return gtk.FALSE

def dimension_horizontal_cb(menuitem, gtkimage):
    _tool = Generic.tools.HorizontalDimensionTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the first point for the dimension.")
    gtkdimension.horizontal_mode_init(_tool)
    return gtk.FALSE

def dimension_vertical_cb(menuitem, gtkimage):
    _tool = Generic.tools.VerticalDimensionTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the first point for the dimension.")
    gtkdimension.vertical_mode_init(_tool)
    return gtk.FALSE

def dimension_radial_cb(menuitem, gtkimage):
    _tool = Generic.tools.RadialDimensionTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on an arc or a circle to dimension.")
    gtkdimension.radial_mode_init(_tool)
    return gtk.FALSE

def dimension_angular_cb(menuitem, gtkimage):
    _tool = Generic.tools.AngularDimensionTool()
    gtkimage.setTool(_tool)
    gtkimage.setPrompt("Click on the angle vertex point or an arc.")
    gtkdimension.angular_mode_init(_tool)
    return gtk.FALSE

def dimension_prefs_cb(menuitem, gtkimage):
    # print "in dimension_prefs_cb()"
    # gtkdimprefs.check_dim_prefs(gtkimage)
    return gtk.FALSE

def debug_menu_init_cb(menuitem, gtkimage):
    return gtk.FALSE

def get_focus_widget_cb(menuitem, gtkimage):
    _widget = gtkimage.getWindow().get_focus()
    print "Focus widget: " + str(_widget)
    return gtk.FALSE

def get_undo_stack_cb(menuitem, gtkimage):
    _layer = gtkimage.getActiveLayer()
    _log = _layer.getLog()
    if _log is not None:
        _log.printUndoData()
    return gtk.FALSE

def get_redo_stack_cb(menuitem, gtkimage):
    _layer = gtkimage.getActiveLayer()
    _log = _layer.getLog()
    if _log is not None:
        _log.printRedoData()
    return gtk.FALSE

def get_image_undo_cb(menuitem, gtkimage):
    gtkimage.printStack(True)
    return gtk.FALSE

def get_image_redo_cb(menuitem, gtkimage):
    gtkimage.printStack(False)
    return gtk.FALSE

def cb(menuitem, window):
    print "called cb"
    return gtk.FALSE

#
# dictionary to bind callbacks
#

callbacks = {
    'file_menu_init' : file_menu_init_cb,
    'file_new' : file_new_cb,
    'file_open' : file_open_cb,
    'file_close' : file_close_cb,
    'file_save' : file_save_cb,
    'file_save_as' : file_save_as_cb,
    'file_save_layer' : file_save_layer_cb,
    'file_quit' : file_quit_cb,
    'edit_menu_init' : edit_menu_init_cb,
    'edit_undo' : edit_undo_cb,
    'edit_redo' : edit_redo_cb,
    'edit_cut' : edit_cut_cb,
    'edit_copy' : edit_copy_cb,
    'edit_paste' : edit_paste_cb,
    'select_cb' : select_cb,
    'select_all_points' : select_all_points_cb,
    'select_all_segments' : select_all_segments_cb,
    'select_all_circles' : select_all_circles_cb,
    'select_all_arcs' : select_all_arcs_cb,
    'select_all_hclines' : select_all_hclines_cb,
    'select_all_vclines' : select_all_vclines_cb,
    'select_all_aclines' : select_all_aclines_cb,
    'select_all_clines' : select_all_clines_cb,
    'select_all_ccircles' : select_all_ccircles_cb,
    'select_all_ldims' : select_all_ldims_cb,
    'select_all_hdims' : select_all_hdims_cb,
    'select_all_vdims' : select_all_vdims_cb,
    'select_all_rdims' : select_all_rdims_cb,
    'select_all_adims' : select_all_adims_cb,
    'prefs' : prefs_cb,
    'draw_menu_init' : draw_menu_init_cb,
    'draw_point' : draw_point_cb,
    'draw_segment' : draw_segment_cb,
    'draw_rectangle' : draw_rectangle_cb,
    'draw_circle_center' : draw_circle_center_cb,
    'draw_circle_tp' : draw_circle_tp_cb,
    'draw_arc_center' : draw_arc_center_cb,
    'draw_hcl' : draw_hcl_cb,
    'draw_vcl' : draw_vcl_cb,
    'draw_acl' : draw_acl_cb,
    'draw_cl' : draw_cl_cb,
    'draw_perpendicular_cline' : draw_perpendicular_cline_cb,
    'draw_tangent_cline' : draw_tangent_cline_cb,
    'draw_tangent_two_ccircles' : draw_tangent_two_ccircles_cb,
    'draw_poffset_cline' : draw_poffset_cline_cb,
    'draw_ccirc_cp' : draw_ccirc_cp_cb,
    'draw_ccirc_tp' : draw_ccirc_tp_cb,
    'draw_tangent_ccircle_menu_init' : draw_tangent_ccircle_menu_init_cb,
    'draw_tangent_single_conobj' : draw_tangent_single_conobj_cb,
    'draw_tangent_two_conobjs' : draw_tangent_two_conobjs_cb,
    'draw_chamfer' : draw_chamfer_cb,
    'draw_fillet' : draw_fillet_cb,
    'draw_leader' : draw_leader_cb,
    'draw_polyline' : draw_polyline_cb,
    'draw_polygon' : draw_polygon_cb,
    'draw_ext_polygon' : draw_ext_polygon_cb,
    'draw_set_style' : draw_set_style,
    'draw_set_linetype' : draw_set_linetype,
    'draw_set_color': draw_set_color,
    'draw_set_thickness' : draw_set_thickness,
    'draw_text' : draw_text_cb,
    'modify_menu_init' : modify_menu_init_cb,
    'transfer' : transfer_object_cb,
    'split' : split_object_cb,
    'mirror' : mirror_object_cb,
    'delete' : delete_cb,
    'zoom' : zoom_cb,
    'zoom_in' : zoom_in_cb,
    'zoom_out' : zoom_out_cb,
    'zoom_fit' : zoom_fit_cb,
    'move_horiz' : move_horizontal_cb,
    'move_vert' : move_vertical_cb,
    'move_twopoint' : move_twopoint_cb,
    'stretch_horiz' : stretch_horiz_cb,
    'stretch_vert' : stretch_vert_cb,
    'stretch_twopoint' : stretch_twopoint_cb,
    'change_style' : change_style_cb,
    'change_color' : change_color_cb,
    'change_linetype' : change_linetype_cb,
    'change_thickness' : change_thickness_cb,
    'dimension_menu_init' : dimension_menu_init_cb,
    'dimension_linear' : dimension_linear_cb,
    'dimension_horiz' : dimension_horizontal_cb,
    'dimension_vert' : dimension_vertical_cb,
    'dimension_rad' : dimension_radial_cb,
    'dimension_ang' : dimension_angular_cb,
    # 'dimension_prefs' : dimension_prefs_cb,
    'debug_menu_init' : debug_menu_init_cb,
    'get_focus_widget' : get_focus_widget_cb,
    'undo_stack' : get_undo_stack_cb,
    'redo_stack' : get_redo_stack_cb,
    'image_undo' : get_image_undo_cb,
    'image_redo' : get_image_redo_cb,
    'cb' : cb
    }

#
# create a Menu object for a list-based menu representation
#

def build_menu(menulist, mdict, gtkimage):
    if not isinstance(menulist, list):
        raise TypeError, "Invalid list for build_menu(): " + `menulist`
    _menu = gtk.Menu()
    for _i in range(len(menulist)):
        _name = menulist[_i][0]
        if _name is None:
            _item = gtk.SeparatorMenuItem()
        else:
            _obj = menulist[_i][1]
            if Generic.menus.submenus.has_key(_obj):
                _cbkey = menulist[_i][2]
                _mlist = Generic.menus.submenus[_obj]
                _item = gtk.MenuItem(_name)
                _submenu = build_menu(_mlist, mdict, gtkimage)
                _item.set_submenu(_submenu)
            else:
                if not callbacks.has_key(_obj):
                    raise KeyError, "Missing callback: %s" % _obj
                _icon, _accel = menulist[_i][2:]
                if _icon is None:
                    _item = gtk.MenuItem(_name)
                else:
                    _item = gtk.ImageMenuItem(_name)
                    if iconmap.has_key(_icon):
                        _image = gtk.image_new_from_stock(iconmap[_icon],
                                                          gtk.ICON_SIZE_MENU)
                        _item.set_image(_image)
                if _accel is not None and accelmap.has_key(_accel):
                    _item.add_accelerator("activate", gtkimage.accel,
                                          accelmap[_accel],
                                          gtk.gdk.CONTROL_MASK,
                                          gtk.ACCEL_VISIBLE)
                #
                # turn off menu items that don't do anything - this
                # is a temporary means of presenting the menus with
                # the items in their positions but make them unselectable
                #
                # as the code to handle them is added, remove them
                # from the test condition ...
                #
                if (_obj == 'cb' or _obj == 'file_save_layer'):
                    _item.set_sensitive(gtk.FALSE)
                else:
                    _item.connect("activate", callbacks[_obj] , gtkimage)
                mdict[id(_item)] = _obj
        _menu.append(_item)
    return _menu

#
# fill a menubar with the menu template info
#

def init_menu(menuitem, gtkimg):
    # print "in init_menu()"
    return gtk.FALSE

def fill_menubar(mb, gtkimage):
    if not isinstance(mb, gtk.MenuBar):
        raise TypeError, "Invalid gtk.MenuBar object: " + `mb`
    _mdict = {}
    for _i in range(len(Generic.menus.primary_menus)):
        _name, _key, _cbkey = Generic.menus.primary_menus[_i]
        assert Generic.menus.submenus.has_key(_key), "Missing primary menu key!"
        if not callbacks.has_key(_cbkey):
            raise ValueError, "Missing menu initializing key: %s" % _cbkey
        _mlist = Generic.menus.submenus[_key]
        _item = gtk.MenuItem(_name)
        _item.connect("activate", callbacks[_cbkey], gtkimage)
        _submenu = build_menu(_mlist, _mdict, gtkimage)
        _item.set_submenu(_submenu)
        mb.append(_item)
    return _mdict



#
# Copyright (c) 2002, Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#
# define __all__ so that "from Interface/Gtk import *" works
#

__all__ = [
    'gtkconobjs',
    'gtkdimension',
    'gtkdimprefs',
    'gtkedit',
    'gtkentities',
    'gtkimage',
    'gtklayers',
    'gtkmenus',
    'gtkmodify',
    'gtkprefs'
    ]



#
# Copyright (c) 2002, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# a class for storing selected objects
#

import image
import layer

class Selection(object):
    """A class used for storing selected objects.

A Selection object has the following methods:

storeObject(): Store a reference to some object.
getObjects(): Return all the objects held within the Selection object.
reset(): Empty the Selection object.

A Selection can have any number of objects added to it, and the
contents of a Selection can be retrieved any number of times.
Once the Selection has had its contents retrieved, however,
the addition of a new object will purge the current contents
of the selection.
    """
    def __init__(self):
        """Initialize a Selection object.

There are no arguments needed for this method.        
        """
        self.__image = None
        self.__layers = {}
        self.__retrieved = False

    def hasObjects(self):
        """Test if there are objects stored in the Selection.

hasObjects()        
        """
        _val = False
        if len(self.__layers):
            for _lid in self.__layers:
                if len(self.__layers[_lid]) > 0:
                    _val = True
                    break
        return _val
    
    def storeObject(self, img, lyr, obj):
        """Store a reference to an object.

storeObject(img, lyr, obj)

The arguments are:

img: An Image object.
lyr: A Layer object.
obj: An object in the drawing (Point, Segment, etc.)
        """
        if not isinstance(img, image.Image):
            raise TypeError, "Invalid image object: " + `img`
        if not isinstance(lyr, layer.Layer):
            raise TypeError, "Invalid layer object: " + `lyr`
        if (self.__retrieved or
            (self.__image is not None and img is not self.__image)):
            self.reset()
        self.__image = img
        _lid = id(lyr)
        _objlist = self.__layers.setdefault(_lid, [])
        _seen = False
        for _i in range(len(_objlist)):
            if obj is objlist[_i]:
                _seen = True
                break
        if not _seen:
            _objlist.append(obj)
        
    def getObjects(self):
        """Return all the currently selected objects.

getObjects()

This method returns a list.
        """
        _allobjs = []
        for _lid in self.__layers:
            _allobjs.extend(self.__layers[_lid][:])
        self.__retrieved = True
        return _allobjs
        
    def reset(self):
        """Reset the Selection object to empty.

reset()        
        """
        self.__image = None
        self.__layers.clear()
        self.__retrieved = False



#
# Copyright (c) 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# quadtree class definition
#
# TODO: Look at using weak references ...

from __future__ import generators

import message

class QTreeNode(message.Messenger):
    messages = {
        'subdivided' : True,
        'reparented' : True,
        'adjusted' : True,
        'full' : True,
        }
    NENODE = 1
    NWNODE = 2
    SWNODE = 3
    SENODE = 4
    def __init__(self):
        message.Messenger.__init__(self)
        self.__subnodes = None
        self.__bounds = None
        self.__parent = None
        self.__threshold = 5
        self.__objects = []

    def __len__(self):
        return len(self.__objects)

    def setBoundary(self, xmin, ymin, xmax, ymax):
        if self.__subnodes is not None or self.__parent is not None:
            raise RuntimeError, "Node cannot have boundary changed."
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        self.__bounds = (_xmin, _ymin, _xmax, _ymax)

    def unsetBoundary(self):
        if self.__subnodes is not None or len(self.__objects) != 0:
            raise RuntimeError, "Node cannot have boundary changed."
        self.__bounds = None

    def getBoundary(self):
        return self.__bounds

    def setThreshold(self, count):
        _t = self.__threshold
        _count = count
        if not isinstance(_count, int):
            _count = int(_count)
        if _count < 1:
            raise ValueError, "Invalid threshold: %d" % _count
        if _count != _t:
            self.__threshold = _count
            self.sendMessage('adjusted', _t)
            if len(self.__objects) >  _t:
                self.sendMessage('full')

    def getThreshold(self):
        return self.__threshold

    def setParent(self, parent):
        _p = self.__parent
        if parent is not None:
            if not isinstance(parent, QTreeNode):
                raise TypeError, "Invalid QTreeNode parent: " + `parent`
        if parent is not _p:
            self.__parent = parent
            self.sendMessage('reparented', _p)

    def getParent(self):
        return self.__parent

    def addObject(self, obj):
        _seen = False
        for _obj in self.__objects:
            if _obj is obj:
                _seen = True
                break
        if not _seen:
            self.__objects.append(obj)
            if len(self.__objects) > self.__threshold:
                self.sendMessage('full')

    def getSubnode(self, subnode):
        if self.__subnodes is None:
            raise RuntimeError, "QTreeNode has no subnodes."
        _ne, _nw, _sw, _se = self.__subnodes
        if subnode == QTreeNode.NENODE:
            _sub = _ne
        elif subnode == QTreeNode.NWNODE:
            _sub = _nw
        elif subnode == QTreeNode.SWNODE:
            _sub = _sw
        elif subnode == QTreeNode.SENODE:
            _sub = _se
        else:
            raise ValueError, "Unknown subnode code: %d" % subnode
        return _sub

    def delObject(self, obj):
        _objs = self.__objects
        for _obj in _objs[:]:
            if obj is _obj:
                _objs.remove(_obj)
                break

    def getObjects(self):
        return self.__objects[:]

    def delObjects(self):
        del self.__objects[:]

    def hasSubnodes(self):
        return self.__subnodes is not None

    def getSubnodes(self):
        return self.__subnodes

    def delSubnodes(self):
        if self.__subnodes is None:
            raise RuntimeError, "QTreeNode has no subnodes."
        for _subnode in self.__subnodes:
            _subnode.setParent(None)
        self.__subnodes = None

    def canSubdivide(self):
        return len(self.__objects) > self.__threshold

    def subdivide(self):
        if len(self.__objects) < self.__threshold:
            return
        _xmin, _ymin, _xmax, _ymax = self.__bounds
        _xmid = (_xmin + _xmax)/2.0
        _ymid = (_ymin + _ymax)/2.0
        #
        # NE Node
        #
        _nenode = QTreeNode()
        _nenode.setBoundary(_xmid, _ymid, _xmax, _ymax)
        _nenode.setThreshold(self.__threshold)
        _nenode.setParent(self)
        #
        # NW Node
        #
        _nwnode = QTreeNode()
        _nwnode.setBoundary(_xmin, _ymid, _xmid, _ymax)
        _nwnode.setThreshold(self.__threshold)
        _nwnode.setParent(self)
        #
        # SW Node
        #
        _swnode = QTreeNode()
        _swnode.setBoundary(_xmin, _ymin, _xmid, _ymid)
        _swnode.setThreshold(self.__threshold)
        _swnode.setParent(self)
        #
        # SE Node
        #
        _senode = QTreeNode()
        _senode.setBoundary(_xmid, _ymin, _xmax, _ymid)
        _senode.setThreshold(self.__threshold)
        _senode.setParent(self)
        #
        self.__subnodes = (_nenode, _nwnode, _swnode, _senode)
        self.delObjects()
        self.sendMessage('subdivided')

    def sendsMessage(self, m):
        if m in QTreeNode.messages:
            return True
        return message.Messenger.sendsMessage(self, m)

class Quadtree(message.Messenger):
    messages = { }
    def __init__(self):
        message.Messenger.__init__(self)
        self.__objects = {}
        self.__queued = []
        self.__split = False
        self.__node = QTreeNode()
        self.__node.connect('full', self, Quadtree.splitTreeNode)

    def __nonzero__(self):
        return len(self.__objects) != 0

    def __len__(self):
        return len(self.__objects)

    def __contains__(self, obj):
        _oid = id(obj)
        return _oid in self.__objects

    def getTreeRoot(self):
        return self.__node

    def getNodes(self, *args):
        _nodes = [self.__node]
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
            else:
                yield _node                
        
    def addObject(self, obj):
        _oid = id(obj)
        if _oid not in self.__objects:
            self.__objects[_oid] = obj

    def getObject(self, obj):
        _oid = id(obj)
        return self.__objects.get(_oid) # returns None if _oid not found

    def getObjects(self):
        return self.__objects.values()

    def delObject(self, obj):
        _oid = id(obj)
        if _oid in self.__objects:
            del self.__objects[_oid]

    def delObjects(self):
        _nodes = [self.__node]
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                _nodes.extend(_node.getSubnodes())
                _node.delSubnodes()
            else:
                for _obj in _node.getObjects():
                    _obj.disconnect(self)
                _node.delObjects()
                if _node.getParent() is not None:
                    _node.setParent(None)
                    _node.disconnect(self)
        self.__objects.clear()
        self.__node.unsetBoundary()

    def queueObject(self, obj):
        _oid = id(obj)
        if _oid in self.__objects:
            raise ValueError, "Object already stored in Quadtree: " + `obj`
        self.__queued.append(obj)

    def emptyQueue(self):
        _objs = self.__queued[:]
        del self.__queued[:]
        return _objs

    def find(self, *params):
        return None

    def getClosest(self, x, y, tol):
        return None

    def getInRegion(self, xmin, ymin, xmax, ymax):
        return []

    def resize(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _root = self.__node
        if not _root.hasSubnodes():
            _root.setBoundary(_xmin, _ymin, _xmax, _ymax)
        else:
            _objs = self.__objects.values()
            self.delObjects()
            _root.setBoundary(_xmin, _ymin, _xmax, _ymax)
            for _obj in _objs:
                self.addObject(_obj)

    def splitTreeNode(self, node, *args):
        if not self.__split:
            _objs = node.getObjects()
            node.subdivide()
            self.__split = True
            for _subnode in node.getSubnodes():
                _subnode.connect('full', self, Quadtree.splitTreeNode)
            for _obj in _objs:
                _oid = id(_obj)
                if _oid in self.__objects:
                    del self.__objects[_oid]
                self.addObject(_obj)
            self.__split = False
                

    def purgeSubnodes(self, node):
        if not isinstance(node, QTreeNode):
            raise TypeError, "Invalid node: " + `node`
        _tmpnode = node
        _parent = node.getParent()
        while _parent is not None:
            _tmpnode = _parent
            _parent = _tmpnode.getParent()
        if self.__node is not _tmpnode:
            raise ValueError, "Node not in tree."
        if node.hasSubnodes():
            _flag = True
            _count = 0
            for _subnode in node.getSubnodes():
                if _subnode.hasSubnodes():
                    _flag = False
                    break
                _count = _count + len(_subnode)
            if _flag and _count < node.getThreshold():
                _objs = []
                for _subnode in node.getSubnodes():
                    _objs.extend(_subnode.getObjects())
                node.delSubnodes()
                for _obj in _objs:
                    Quadtree.delObject(self, _obj)
                    self.addObject(_obj)

    def sendsMessage(self, m):
        if m in Quadtree.messages:
            return True
        return message.Messenger.sendsMessage(self, m)



#
# This file stores the global PythonCAD preferences, and
# is written itself in Python. Copy this file to the file
# '/etc/pythoncad/prefs.py' and it will be loaded and used
# to store global preferences for the program. Whenever
# a new image is opened, that image will use the global
# preferences for setting itself up. Each drawing window
# can then adjust these preferences by using the Preferences
# menus.
#
# When PythonCAD starts, a set of default values for the
# various options any image may set is stored, then the
# '/etc/pythoncad/prefs.py' file is searched for. If it
# is found, the values in the file will override the initial
# set of defaults. Variables defined in this file can
# therefore be commented out without problem.
#
# Any user can copy this file to their home directory and
# store it in a '.pythoncad' directory if they choose. By
# modifying their local preference file the user can add
# new linestyles, new dimension styles, or tailor the
# default configuration of the program to whatever works
# best.
#
# When defining a string value, the string can be in
# any case: 'UPPER', 'lower', or 'MiXeD'. The program will
# convert the string to upper case when testing the values
# entered here. Any time a string is given but the program
# fails to accept it due to case issues is a bug ...
#
# User Prefs
#
# This variable is meant as an adminstrators tool of
# disabling the reading of a preference file in the
# user's home directory. By default PythonCAD will look
# for a file in ${HOME}/.pythoncad/prefs.py after reading
# the global preference file '/etc/pythoncad/prefs.py'.
# If the user_prefs variable is set to False, the search
# for a user's preference file is not done.
#
# The prescence of this variable only makes sense in
# the '/etc/pythoncad.prefs' file.
#
# Valid choices: True, False

user_prefs = True

#
# Units
#
# The basic unit of length in an image.
#
# Valid choices:
# 'millimeters', 'micrometers', 'meters', 'kilometers',
# 'inches', 'feet', 'yards', 'miles'
#

units = 'millimeters'

#
# Chamfer Length
#
# The default chamfer length in an image.
#
# Valid values: Any numerical value 0.0 or greater
#

chamfer_length = 1.5

#
# Fillet Radius
#
# The default fillet radius in an image.
#
# Valid values: Any numerical value 0.0 or greater
#

fillet_radius = 2.0

#
# Leader arrow size
#
# This option sets the default arrow size on leader lines.
#
# Valid values: Any numerical value 0.0 or greater
#

leader_arrow_size = 10.0

#
# Dimension Options
#

#
# Dual Dimension display
#
# This option will set any new dimension object to display
# both possible dimension strings by default.
#
# Valid choices: True, False
#

dim_dual_mode = True

#
# Highlight Points
#
# This option will activate the drawing of small boxes around
# the Point entities in an image.
#
# Valid choices: True, False
#

highlight_points = True

#
# Linetypes
#
# There are several default Linetypes provided in PythonCAD.
# If you want to define more, they are added here.
#
# Each Linetype is defined as a tuple of two objects:
#
# (name, dashlist)
#
# The name is a string or unicode string, and the dash list
# is either 'None' (meaning a solid line), or a list with
# an even number of integer values. Each value pair represents
# the on-off bit pattern in the line.
#
# Examples:
# A dashed linetype called 'dash' with 4 bits on then 4 bits off:
# ('dash', [4, 4])
#
# A dashed linetype called 'dash2' with 10 bits on, 2 off, 6 on, 2 off
# (u'dash2', [10, 2, 6, 2]) # the "u" means unicode ...
#
# Add any new linetypes you wish in the linetypes list variable
#

linetypes = [
    # ('dash', [4, 4]), # dash example above
    # (u'dash2', [10, 2, 6, 2]), # dash2 example above
    ]

#
# Colors
#
# By default PythonCAD starts with eight basic colors defined.
# If you want to add more, here is where to add them.
#
# A Color is defined in one of two ways:
#
# '#xxxxxx' - a hexidecimal string like '#ff00ff'
# (r, g, b) - a tuple with three integers giving the red, green, and
#             blue values. Each value must be 0 <= val <= 255.
#
# Add any new colors you wish to use in the colors list variable

colors = [
    # '#ffffff', # hexidecimal example
    # '#33cc77', # another hex example
    # (100, 50, 30), # tuple example
    # (255, 30, 180), # another tuple example
    ]

#
# Styles
#
# Styles define a common set of properties for objects like
# segments, circles, and arcs. There are eight default styles
# in PythonCAD, and adding more styles can be done here.
#
# Defining a style is done by creating a tuple containing
# four fields:
#
# (name, linetype, color, thickness)
#
# The 'name' field is simply a string or unicode string. The
# 'linetype' field is a two object tuple described above, and
# the 'color' field is described above also. The 'thickness'
# field is a positive float value giving the line thickness.
#
# Examples:
#
# (u'style1', (u'lt1', [10, 4]), '#ffcc99', 0.1)
# A style called 'style1', composed of a linetype called 'lt1'
# which is a dashed line 10 bits on, 4 off, drawn with the
# color '#ffcc99' and 0.1 units thick.
#
# Add any new styles you want to use to the 'styles' list
# variable below. Several examples are included to provide
# a starting guide for creating new styles.
#
#

styles = [
    # (u'style1', (u'lt1', [10, 4]), '#ffcc99', 0.1), # example
    # (u'style2', (u'lt2', [6, 2, 10, 4]), '#cc00ff', 0.3), # another example
    # (u'style3', (u'lt3', [2, 6, 2, 2]), '#8844bb', 0.2), # third example
    ]

#
# Default style
#
# Set this variable to the name of the style you want the
# initial drawing style to be in PythonCAD. If you set this
# variable to None, or comment it out, the default style
# will be a solid white line.

default_style = None

#
# Dimension Styles
#
# A dimension style is a set of parameters that define
# how the dimension looks. Any dimension has a large number
# of attributes that can be set, so a dimension style is
# a way of grouping together these attributes into a common
# set and offering a means of making the dimensions look
# uniform.
#
# Defining a dimension style requires creating a dictionary
# with a certain keys and values. PythonCAD defines a single
# default dimension style, and any new dimension styles that
# are created here use that default style as a base, and the
# values in the new style then override the defaults.
#
# The list of dictionary keys and acceptable values is
# somewhat long. This list is given below with each key
# and the acceptable values for each key listed.
#
# Key: 'DIM_PRIMARY_FONT_FAMILY'
# Values: A string giving the font name
# Default: 'Sans'
#
# Key: 'DIM_PRIMARY_FONT_SIZE'
# Values: A positive integer value
# Default: 12
#
# Key: 'DIM_PRIMARY_FONT_WEIGHT'
# Values: Either 'Normal', 'Light', 'Bold', or 'Heavy'
# Default: 'Normal'
#
# Key: 'DIM_PRIMARY_FONT_STYLE'
# Values: 'Normal', 'Oblique', or 'Italic'
# Default: 'Normal'
#
# Key: 'DIM_PRIMARY_FONT_COLOR'
# Values: A Color definition (see above), like (255, 255, 255) or '#ffffff'
# Default: '#ffffff'
#
# Key: 'DIM_PRIMARY_PREFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'DIM_PRIMARY_SUFFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'DIM_PRIMARY_PRECISION'
# Value: An integer value where 0 <= value <= 15
# Default: 3
#
# Key: 'DIM_PRIMARY_UNITS'
# Value: 'Millimeters', 'Micrometers', 'Meters', 'Milometers',
#        'Inches', 'Feet', 'Yards', 'Miles'
# Default: 'Millimeters'
#
# Key: 'DIM_PRIMARY_LEADING_ZERO'
# Value: True or False
# Default: True
#
# Key: 'DIM_PRIMARY_TRAILING_DECIMAL'
# Value: True or False
# Default: True
#
# Key: 'DIM_SECONDARY_FONT_FAMILY'
# Values: A string giving the font name
# Default: 'Sans'
#
# Key: 'DIM_SECONDARY_FONT_SIZE'
# Values: A positive integer value
# Default: 12
#
# Key: 'DIM_SECONDARY_FONT_WEIGHT'
# Values: Either 'Normal', 'Light', 'Bold', or 'Heavy'
# Default: 'Normal'
#
# Key: 'DIM_SECONDARY_FONT_STYLE'
# Values: 'Normal', 'Oblique', or 'Italic'
# Default: 'Normal'
#
# Key: 'DIM_SECONDARY_FONT_COLOR'
# Values: A Color definition (see above), like (255, 255, 255) or '#ffffff'
# Default: '#ffffff'
#
# Key: 'DIM_SECONDARY_PREFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'DIM_SECONDARY_SUFFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'DIM_SECONDARY_PRECISION'
# Value: An integer value where 0 <= value <= 15
# Default: u''
#
# Key: 'DIM_SECONDARY_UNITS'
# Value: 'Millimeters', 'Micrometers', 'Meters', 'Milometers',
#        'Inches', 'Feet', 'Yards', 'Miles'
# Default: 'Millimeters'
#
# Key: 'DIM_SECONDARY_LEADING_ZERO'
# Value: True or False
# Default: True
#
# Key: 'DIM_SECONDARY_TRAILING_DECIMAL'
# Value: True or False
# Default: True
#
# Key: 'DIM_OFFSET'
# Value: A float value of 0.0 or greater.
# Default: 1.0
#
# Key: 'DIM_EXTENSION'
# Value: A float value of 0.0 or greater.
# Default: 1.0
#
# Key: 'DIM_COLOR'
# Value: A Color definition (see above), like (255, 255, 255) or '#ffffff'
# Default: (255, 165, 0)
#
# Key: 'DIM_POSITION'
# Value: 'Split', 'Above', 'Below'
# Default: 'Split'
#
# Key: 'DIM_ENDPOINT'
# Value: 'No_endpoint' or 'None' or None, 'Arrow', 'Filled_arrow',
#        'Slash', 'Circle'
# Default: None
#
# Key: 'DIM_ENDPOINT_SIZE'
# Value: A float vale of 0.0 or greater.
# Default: 1.0
#
# Key: 'DIM_DUAL_MODE'
# Value: True or False
# Default: False
#
# Key: 'RADIAL_DIM_PRIMARY_PREFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'RADIAL_DIM_PRIMARY_SUFFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'RADIAL_DIM_SECONDARY_PREFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'RADIAL_DIM_SECONDARY_SUFFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'RADIAL_DIM_DIA_MODE'
# Value: True or False
# Default: False
#
# Key: 'ANGULAR_DIM_PRIMARY_PREFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'ANGULAR_DIM_PRIMARY_SUFFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'ANGULAR_DIM_SECONDARY_PREFIX'
# Value: A string or unicode string
# Default: u''
#
# Key: 'ANGULAR_DIM_SECONDARY_SUFFIX'
# Value: A string or unicode string
# Default: u''
#
#
# The keys that begin 'DIM_PRIMARY_' apply are specific to the
# primary dimension in any dimension, and 'DIM_SECONDARY_' are
# specific to the secondary dimension. 'RADIAL_DIM_' and
# 'ANGULAR_DIM_' are obviously applicible to those specific
# dimension types.
#
# 'DIM_OFFSET' is the distance between the dimensioned point and
# the start of the dimension bar, and 'DIM_EXTENSION' is the
# distance between where the dimension crossbar intersects the
# dimension bar and the end of the dimension bar. 'DIM_POSITION'
# will be used in placing the dimension text around the dimension
# crossbar. The names of the keys were chosen to be somewhat
# self explanatory, but they could change in future releases.
#
# As a dimension style is built by overriding the default values,
# creating a new style only requires changing the fields that
# you wish to update. In the example below, the four fields
# 'dim_offset', 'dim_endpoint', 'dim_endpoint_size', and
# 'dim_extension' are modified.

example_dict = {
    'dim_offset' : 5.0,
    'dim_endpoint' : 'filled_arrow',
    'dim_endpoint_size' : 10.0,
    'dim_extension' : 5.0,
    }
#
# here is an example of a dimension style with more changes
#

dimstyle_example_2 = {
    'dim_primary_font_family' : 'Times',
    'dim_primary_font_size' : 24,
    'dim_primary_font_color' : '#dd44bb',
    'dim_primary_suffix' : u' mm',
    'dim_primary_precision' : 2,
    'dim_secondary_font_family' : 'Helvetica',
    'dim_secondary_font_size' : 18,
    'dim_secondary_font_color' : '#33ee55',
    'dim_secondary_suffix' : u' in.',
    'dim_secondary_units' : 'inches',
    'dim_offset' : 10.0,
    'dim_endpoint' : 'slash',
    'dim_endpoint_size' : 10.0,
    'dim_extension' : 5.0,
    'radial_dim_primary_prefix' : u'R. ',
    'radial_dim_primary_suffix' : u' mm',
    'radial_dim_secondary_prefix' : u'R. ',
    'radial_dim_secondary_suffix' : u' in.',
    'dim_dual_mode' : True,
    }
#
# The creation of the dimension style is done with by creating
# a tuple containing two objects:
#
# (name, dict)
#
# name: The dimension style name
# dict: The dictionary containing the values for the dimension style
#
# Add any dimension styles you create to the dimstyles list below.

dimstyles = [
    # (u'example_dimstyle', example_dict), # example
    # (u'ex2' , dimstyle_example_2), # another example
    ]

#
# Default dimension style
#
# Set this variable to the name of the dimension style you
# want PythonCAD to use by default. If you define this variable
# to None, or comment it out, the default dimension style
# is used.

default_dimstyle = None

#
# Text styles
#
# A text style defines a common set of text attributes
# that are used to define visual elements of some text.
# A single text style is defined in PythonCAD, so if you
# want more text styles here is where to add them.
#
# Defining a text style is done by creating a tuple containing
# six fields:
#
# (name, family, size, style, weight, color)
#
# The 'name' field is the name of the text style, and can be
# used to reference it. The 'family' field is the font family,
# like 'Sans' or 'Times'. The 'size' field is the text size in
# points. The style can be either 'normal, 'oblique', or 'italic'.
# The 'weight' can be one of 'normal', 'light', 'bold', or 'heavy',
# and the color is a color definition (see above).
#
# Examples:
#
# ('default', 'sans', 10, 'normal', 'normal', '#ff00cc')
# A simple 10-point Sans format of color #ff00cc called 'default'
#
# ('fancy', 'helvetica', 20, 'italic', 'bold', '#ffff00')
# 20-point bold, italicized Helvetica of color #ffff00 called 'fancy'.
#
# Defining fonts style (and font handling) is still in the
# earliest stages in PythonCAD, so the sections dealing with
# text will most like change ...
#
# Add more text styles you wish to use here
#

textstyles = [
    # ('default', 'sans', 10, 'normal', 'normal', '#ff00cc') # example
    ]

#
# Default text style
#
# Set this variable to the name of the text style you
# want PythonCAD to use by default. If you define this variable
# to None, or comment it out, the default text style is used
#

default_textstyle = None

#
# Text properties
#
# In addition to setting the text styles, some default font properties
# can be set in this file. In setting one of the following values, the
# value defined in the default textstyle will be overriden.
#
# The examples below are all commented out. Uncomment the ones you
# want to override the default style info.
#
# Default font family
#
# font_family = 'Sans' # uncomment this to set the default font family
#
# Default font size
# This value must be an integer greater than 0.
#
# font_size = 18 # uncomment this to set the default font size
#
# Default font weight
# This can be 'normal', 'light', 'bold', or 'heavy'.
#
# font_weight = 'normal' # uncomment this to set the default font weight
#
# Default font style
# This can be 'normal', 'italic', or 'oblique'.
#
# font_style = 'normal' # uncomment this to set the default font style
#
# Default font color
# This must be a color definition.
#
# font_color = '#ff00ff' # uncomment this to set the default font color

#
# Miscellaneous options
#
# Autosplitting (NOT IMPLEMENTED YET)
#
# This option will activate the automatic splitting of segments,
# circles, and arcs if a new point is added directly on the object.
#
# Valid choices: True, False
#

autosplit = False



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# vertical construction lines
#

from __future__ import generators

import conobject
import entity
import tolerance
import point
import quadtree
import logger

class VCLine(conobject.ConstructionObject):
    """A base class for horizontal construction lines.
    """
    messages = {
        'moved' : True,
        'keypoint_changed' : True
    }
    def __init__(self, p, **kw):
        """Instantiate an VCLine object.

VCLine(p)

        """
        _p = p
        if not isinstance(p, point.Point):
            _p = point.Point(p)
        conobject.ConstructionObject.__init__(self, **kw)
        self.__keypoint = _p
        _p.storeUser(self)
        _p.connect('moved', self, VCLine.movePoint)

    def __eq__(self, obj):
        """Compare one VCLine to another for equality.
        """
        if not isinstance(obj, VCLine):
            return False
        if obj is self:
            return True
        if abs(self.getLocation().x - obj.getLocation().x) < 1e-10:
            return True
        return False

    def __ne__(self, obj):
        """Compare one VCLine to another for inequality.
        """
        if not isinstance(obj, VCLine):
            return True
        if obj is self:
            return False
        if abs(self.getLocation().x - obj.getLocation().x) < 1e-10:
            return False
        return True

    def __str__(self):
        _x, _y = self.getLocation().getCoords()
        return "Vertical Construction Line at x = %g" % self.__keypoint.x

    def getValues(self):
        return 'vcline', entity.Entity.getValues(self), self.__keypoint.getID()

    def getLocation(self):
        return self.__keypoint

    def setLocation(self, p):
        if self.isLocked():
            raise RuntimeError, "Setting keypoint not allowed - object locked."
        if not isinstance(p, point.Point):
            raise TypeError, "Invalid keypoint: " + `p`
        _kp = self.__keypoint
        if p is not _kp:
            _x = _kp.x
            _kp.disconnect(self)
            _kp.freeUser(self)
            self.__keypoint = p
            self.sendMessage('keypoint_changed', _kp)
            p.connect('moved', self, VCLine.movePoint)
            p.storeUser(self)
            if abs(_x - p.x) > 1e-10:
                self.sendMessage('moved', _x, p.y)
            self.sendMessage('modified')

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest Point on the VCLine to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required argument:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual Point on the VCLine. If the distance between the actual
Point and the coordinates used as an argument is less than the tolerance,
the actual Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _vx = self.__keypoint.x
        if abs(_vx - x) < _t:
            return _vx, _y
        return None

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not an VCLine passes through a region.

inRegion(xmin, ymin, xmax, ymax)

The first four arguments define the boundary. The method
will return True if the VCLine falls between xmin and xmax.
Otherwise the function will return False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + str(fully)
        if fully:
            return False
        _x = self.__keypoint.x
        return _xmin < _x < _xmax

    def move(self, dx, dy):
        """Move a VCLine

move(dx, dy)

The first argument gives the x-coordinate displacement,
and the second gives the y-coordinate displacement. Both
values should be floats.
        """
        if self.isLocked() or self.__keypoint.isLocked():
            raise RuntimeError, "Moving VCLine not allowed - object locked."
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        if abs(_dx) > 1e-10 or abs(_dy) > 1e-10:
            _x, _y = self.__keypoint.getCoords()
            self.ignore('moved')
            self.__keypoint.move(_dx, _dy)
            self.receive('moved')
            self.sendMessage('moved', _x, _y)
            self.modified()

    def movePoint(self, p, *args):
        _plen = len(args)
        if _plen < 2:
            raise ValueError, "Invalid argument count: %d" % _plen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        if p is not self.__keypoint:
            raise ValueError, "Invalid point for VCLine::movePoint()" + `p`
        if abs(p.x - _x) > 1e-10:
            self.sendMessage('moved', _x, _y)
            self.modified()

    def clipToRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _x = self.__keypoint.x
        if _xmin < _x < _xmax:
            return _x, _ymin, _x, _ymax
        return None

    def sendsMessage(self, m):
        if m in VCLine.messages:
            return True
        return conobject.ConstructionObject.sendsMessage(self, m)

def intersect_region(vcl, xmin, ymin, xmax, ymax):
    if not isinstance(vcl, VCLine):
        raise TypeError, "Invalid VCLine: " + str(vcl)
    _xmin = xmin
    if not isinstance(_xmin, float):
        _xmin = float(xmin)
    _ymin = ymin
    if not isinstance(_ymin, float):
        _ymin = float(ymin)
    _xmax = xmax
    if not isinstance(_xmax, float):
        _xmax = float(xmax)
    if _xmax < _xmin:
        raise ValueError, "Illegal values: xmax < xmin"
    _ymax = ymax
    if not isinstance(_ymax, float):
        _ymax = float(ymax)
    if _ymax < _ymin:
        raise ValueError, "Illegal values: ymax < ymin"
    _x, _y = vcl.getLocation().getCoords()
    _x1 = _y1 = _x2 = _y2 = None
    if _xmin < _x < _xmax:
        _x1 = _x
        _y1 = _ymin
        _x2 = _x
        _y2 = _ymax
    return _x1, _y1, _x2, _y2

#
# Quadtree VCLine storage
#

class VCLineQuadtree(quadtree.Quadtree):
    def __init__(self):
        quadtree.Quadtree.__init__(self)

    def getNodes(self, *args):
        _alen = len(args)
        if _alen != 1:
            raise ValueError, "Expected 1 arguments, got %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _nodes = [self.getTreeRoot()]
        while len(_nodes):
            _node = _nodes.pop()
            _bounds = _node.getBoundary()
            _xmin = _bounds[0]
            _xmax = _bounds[2]
            if _x < _xmin or _x > _xmax:
                continue
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ne = _nw = _sw = _se = True
                if _x < _xmid: # vcline to left
                    _ne = _se = False
                else: # vcline to right
                    _nw = _sw = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                yield _node
      
    def addObject(self, obj):
        if not isinstance(obj, VCLine):
            raise TypeError, "Invalid VCLine object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getLocation().getCoords()
        _bounds = self.getTreeRoot().getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _x - 1.0
            _ymin = _y - 1.0
            _xmax = _x + 1.0
            _ymax = _y + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _x < _xmin:
                _xmin = _x - 1.0
                _resize = True
            if _x > _xmax:
                _xmax = _x + 1.0
                _resize = True
            if _y < _ymin:
                _ymin = _y - 1.0
                _resize = True
            if _y > _ymax:
                _ymax = _y + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
        for _node in self.getNodes(_x):
            _xmin, _ymin, _xmax, _ymax = _node.getBoundary()            
            if obj.inRegion(_xmin, _ymin, _xmax, _ymax):
                _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, VCLineQuadtree.moveVCLine)

    def old_addObject(self, obj):
        if not isinstance(obj, VCLine):
            raise TypeError, "Invalid VCLine object: " + `obj`
        if obj in self:
            return
        _x, _y = obj.getLocation().getCoords()
        _node = self.getTreeRoot()
        _bounds = _node.getBoundary()
        _xmin = _ymin = _xmax = _ymax = None
        _resize = False
        if _bounds is None: # first node in tree
            _resize = True
            _xmin = _x - 1.0
            _ymin = _y - 1.0
            _xmax = _x + 1.0
            _ymax = _y + 1.0
        else:
            _xmin, _ymin, _xmax, _ymax = _bounds
            if _x < _xmin:
                _xmin = _x - 1.0
                _resize = True
            if _x > _xmax:
                _xmax = _x + 1.0
                _resize = True
            if _y < _ymin:
                _ymin = _y - 1.0
                _resize = True
            if _y > _ymax:
                _ymax = _y + 1.0
                _resize = True
        if _resize:
            self.resize(_xmin, _ymin, _xmax, _ymax)
            _node = self.getTreeRoot()
        _nodes = []
        _nodes.append(_node)
        while len(_nodes):
            _node = _nodes.pop()
            _bounds = _node.getBoundary()
            _xmin = _bounds[0]
            _xmax = _bounds[2]
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ne = _nw = _sw = _se = True
                if _x < _xmid: # vcline to left
                    _ne = _se = False
                else: # vcline to right
                    _nw = _sw = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if (not (_x < _xmin) and (_x < _xmax)):
                    _node.addObject(obj)
        quadtree.Quadtree.addObject(self, obj)
        obj.connect('moved', self, VCLineQuadtree.moveVCLine)

    def delObject(self, obj):
        if obj not in self:
            return
        _pdict = {}
        _x, _y = obj.getLocation().getCoords()
        for _node in self.getNodes(_x):
            _node.delObject(obj)
            _parent = _node.getParent()
            if _parent is not None:
                _pid = id(_parent)
                if _pid not in _pdict:
                    _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def old_delObject(self, obj):
        if obj not in self:
            return
        _x, _y = obj.getLocation().getCoords()
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _pdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            _bounds = _node.getBoundary()
            _xmin = _bounds[0]
            _xmax = _bounds[2]
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ne = _nw = _sw = _se = True
                if _x < _xmid: # vcline to left
                    _ne = _se = False
                else: # vcline to right
                    _nw = _sw = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                if (not (_x < _xmin) and (_x < _xmax)):
                    _node.delObject(obj)
                    _parent = _node.getParent()
                    if _parent is not None:
                        _pid = id(_parent)
                        if _pid not in _pdict:
                            _pdict[_pid] = _parent
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        for _parent in _pdict.values():
            self.purgeSubnodes(_parent)

    def find(self, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _t = tolerance.TOL
        if _alen > 1:
            _t = tolerance.toltest(args[1])
        if not len(self):
            return None
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _vcl = _tsep = None
        _vdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            _bounds = _node.getBoundary()
            _xmin = _bounds[0]
            _xmax = _bounds[2]
            if (_xmin - _t) < _x < (_xmax + _t):
                if _node.hasSubnodes():
                    for _subnode in _node.getSubnodes():
                        _bounds = _node.getBoundary()
                        _xmin = _bounds[0]
                        _xmax = _bounds[2]
                        if (_xmin - _t) < _x < (_xmax + _t):
                            _nodes.append(_subnode)
                else:
                    for _v in _node.getObjects():
                        _vid = id(_v)
                        if _vid not in _vdict:
                            _xv = _v.getLocation().x
                            _xsep = abs(_xv - _x)
                            if _xsep < _t:
                                if _vcl is None:
                                    _vcl = _v
                                    _tsep = _xsep
                                else:
                                    if _xsep < _tsep:
                                        _vcl = _v
                                        _tsep = _xsep
                            _vdict[_vid] = True
        return _vcl

    def moveVCLine(self, obj, *args):
        if obj not in self:
            raise ValueError, "VCLine not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        for _node in self.getNodes(_x):
            _node.delObject(obj) # vcline may not be in node
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def old_moveVCLine(self, obj, *args):
        if obj not in self:
            raise ValueError, "VCLine not stored in Quadtree: " + `obj`
        _alen = len(args)
        if _alen < 2:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x = args[0]
        if not isinstance(_x, float):
            _x = float(args[0])
        _y = args[1]
        if not isinstance(_y, float):
            _y = float(args[1])
        _nodes = []
        _nodes.append(self.getTreeRoot())
        while len(_nodes):
            _node = _nodes.pop()
            _bounds = _node.getBoundary()
            _xmin = _bounds[0]
            _xmax = _bounds[2]
            if _node.hasSubnodes():
                _xmid = (_xmin + _xmax)/2.0
                _ne = _nw = _sw = _se = True
                if _x < _xmid: # vcline to left
                    _ne = _se = False
                else: # vcline to right
                    _nw = _sw = False
                if _ne:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NENODE))
                if _nw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.NWNODE))
                if _sw:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SWNODE))
                if _se:
                    _nodes.append(_node.getSubnode(quadtree.QTreeNode.SENODE))
            else:
                _node.delObject(obj) # vcline may not be in node
        quadtree.Quadtree.delObject(self, obj)
        obj.disconnect(self)
        self.addObject(obj)

    def getClosest(self, x, y, tol=tolerance.TOL):
        return self.find(x, tol)

    def getInRegion(self, xmin, ymin, xmax, ymax):
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        _vcls = []
        if not len(self):
            return _vcls
        _nodes = []
        _nodes.append(self.getTreeRoot())
        _vdict = {}
        while len(_nodes):
            _node = _nodes.pop()
            if _node.hasSubnodes():
                for _subnode in _node.getSubnodes():
                    _bounds = _subnode.getBoundary()
                    _bmin = _bounds[0]
                    _bmax = _bounds[2]
                    if ((_bmin > _xmax) or (_bmax < _xmin)):
                        continue
                    _nodes.append(_subnode)
            else:
                for _vcl in _node.getObjects():
                    _vid = id(_vcl)
                    if _vid not in _vdict:
                        if _vcl.inRegion(_xmin, _ymin, _xmax, _ymax):
                            _vcls.append(_vcl)
                        _vdict[_vid] = True
        return _vcls

#
# VCLine history class
#

class VCLineLog(logger.Logger):
    def __init__(self, v):
        if not isinstance(v, VCLine):
            raise TypeError, "Invalid VCLine: " + `v`
        logger.Logger.__init__(self)
        self.__v = v
        v.connect('keypoint_changed' ,self, VCLineLog.keypointChange)

    def detatch(self):
        self.__v.disconnect(self)
        self.__v = None

    def keypointChange(self, v, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _old = args[0]
        if not isinstance(_old, point.Point):
            raise TypeError, "Invalid old endpoint: " + `_old`
        _oid = _old.getID()
        self.saveUndoData('keypoint_changed', _oid)

    def execute(self, undo, *args):
        if undo is not True and undo is not False:
            raise ValueError, "Invalid undo value: " + str(undo)
        _alen = len(args)
        if _alen == 0:
            raise ValueError, "No arguments to execute()"
        _v = self.__v
        _p = _v.getLocation()
        _op = args[0]
        if _op == 'keypoint_changed':
            if _alen < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _oid = args[1]
            _cid = _p.getID()
            if undo:
                self.saveRedoData('keypoint_changed', _cid)
                self.ignore('keypoint_changed')
                try:
                    _parent = _v.getParent()
                    if _parent is None:
                        raise ValueError, "VCLine has no parent - cannot undo"
                    _op = _parent.getObject(_oid)
                    if _op is None or not isinstance(_op, point.Point):
                        raise ValueError, "Keypoint missing: id=%d" % _oid
                    _x1, _y1 = _p.getCoords()
                    _v.startUndo(True)
                    try:
                        _v.setLocation(_op)
                    finally:
                        _v.endUndo()
                    if abs(_op.x - _x1) > 1e-10:
                        _v.sendMessage('moved', _x1, _y1)
                finally:
                    self.receive('keypoint_changed')
            else:
                self.saveUndoData('keypoint_changed', _cid)
                _parent = _v.getParent()
                if _parent is None:
                    raise ValueError, "VCLine has no parent - cannot undo"
                _op = _parent.getObject(_oid)
                if _op is None or not isinstance(_op, point.Point):
                    raise ValueError, "Keypoint missing: id=%d" % _oid
                self.ignore('keypoint_changed')
                try:
                    _v.setLocation(_op)
                finally:
                    self.receive('keypoint_changed')
        else:
            raise ValueError, "Unexpected operation: %s" % _op
